from .pdistributions import Prior_distribution
from .uniform import Uniform_prior
from .normal import Normal_prior
from .gen_normal import Gen_normal_prior
from .gamma import Gamma_prior
from .invgamma import Invgamma_prior
from .update_pdis import update_pdis

__all__ = ["Prior_distribution","Uniform_prior","Normal_prior","Gen_normal_prior","Gamma_prior","Invgamma_prior","update_pdis"]



import numpy as np
from .pdistributions import Prior_distribution

class Uniform_prior(Prior_distribution):
    def __init__(self,start=-18.0,end=18.0,prob=1.0,**kwargs):
        """ 
        Uniform prior distribution used for each type of hyperparameters in log-space. 
        If the type of the hyperparameter is multi dimensional (H) it is given in the axis=-1. 
        If multiple values (M) of the hyperparameter(/s) are calculated simultaneously it has to be in a (M,H) array. 
        Parameters:
            start: float or (H) array
                The start of non-zero prior distribution value of the hyperparameter in log-space. 
            end: float or (H) array
                The end of non-zero prior distribution value of the hyperparameter in log-space. 
            prob: float or (H) array
                The non-zero prior distribution value.
        """
        self.update_arguments(start=start,
                              end=end,
                              prob=prob,
                              **kwargs)
    
    def ln_pdf(self,x):
        ln_0=-np.log(np.nan_to_num(np.inf))
        if self.nosum:
            return np.where(x>=self.start,np.where(x<=self.end,np.log(self.prob),ln_0),ln_0)
        return np.sum(np.where(x>=self.start,np.where(x<=self.end,np.log(self.prob),ln_0),ln_0),axis=-1)
    
    def ln_deriv(self,x):
        return 0.0*x
    
    def update_arguments(self,start=None,end=None,prob=None,**kwargs):
        """
        Update the object with its arguments. The existing arguments are used if they are not given.
        Parameters:
            start: float or (H) array
                The start of non-zero prior distribution value of the hyperparameter in log-space. 
            end: float or (H) array
                The end of non-zero prior distribution value of the hyperparameter in log-space. 
            prob: float or (H) array
                The non-zero prior distribution value.
        Returns:
            self: The updated object itself.
        """
        if start is not None:
            if isinstance(start,(float,int)):
                self.start=start
            else:
                self.start=np.array(start).reshape(-1)
        if end is not None:
            if isinstance(end,(float,int)):
                self.end=end
            else:
                self.end=np.array(end).reshape(-1)
        if prob is not None:
            if isinstance(prob,(float,int)):
                self.prob=prob
            else:
                self.prob=np.array(prob).reshape(-1)
        if isinstance(self.start,(float,int)) and isinstance(self.end,(float,int)) and isinstance(self.prob,(float,int)):
            self.nosum=True
        else:
            self.nosum=False
        return self
            
    def mean_var(self,mean,var):
        std=np.sqrt(var)
        return self.update_arguments(start=mean-4.0*std,end=mean+4.0*std,prob=1.0/(8.0*std))
    
    def min_max(self,min_v,max_v):
        return self.update_arguments(start=min_v,end=max_v,prob=1.0/(max_v-min_v))
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(start=self.start,end=self.start,prob=self.prob)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs



from .optimizer import Optimizer
import numpy as np

class GlobalOptimizer(Optimizer):
    def __init__(self,local_optimizer=None,bounds=None,maxiter=5000,**kwargs):
        """
        The global optimizer used for optimzing the objective function wrt. the hyperparameters.
        The global optimizer requires a local optimization method and boundary conditions of the hyperparameters.
        Parameters:
            local_optimizer : Local optimizer class
                A local optimization method.
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
            maxiter : int
                The maximum number of evaluations or iterations the global optimizer can use.
        """
        # This global optimizer can not be parallelized
        self.parallel=False
        # The gradient of the objective function is not used by the global optimizer
        self.jac=False
        # Set default bounds
        if bounds is None:
            from ..hpboundary.hptrans import VariableTransformation
            bounds=VariableTransformation(bounds=None)
        # Set default local optimizer
        if local_optimizer is None:
            from .localoptimizer import ScipyOptimizer
            local_optimizer=ScipyOptimizer(maxiter=maxiter,bounds=bounds,use_bounds=False)
        # Set all the arguments
        self.update_arguments(local_optimizer=local_optimizer,
                              bounds=bounds,
                              maxiter=maxiter,
                              **kwargs)

    def run(self,func,theta,parameters,model,X,Y,pdis,**kwargs):
        raise NotImplementedError()

    def update_arguments(self,local_optimizer=None,bounds=None,maxiter=None,**kwargs):
        """
        Update the optimizer with its arguments. The existing arguments are used if they are not given.
        Parameters:
            local_optimizer : Local optimizer class
                A local optimization method.
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
        Returns:
            self: The updated object itself.
        """
        if local_optimizer is not None:
            self.local_optimizer=local_optimizer.copy()
        if bounds is not None:
            self.bounds=bounds.copy()
            # Use the same boundary conditions in the local optimizer
            self.local_optimizer.update_arguments(bounds=self.bounds)
        if maxiter is not None:
            self.maxiter=int(maxiter)
        return self
    
    def run_local_opt(self,func,theta,parameters,model,X,Y,pdis,**kwargs):
        " Run the local optimization. "
        return self.local_optimizer.run(func,theta,parameters,model,X,Y,pdis,**kwargs)
    
    def make_lines(self,parameters,ngrid,**kwargs):
        " Make the lines of the hyperparameters from boundary conditions. "
        return self.bounds.make_lines(parameters=parameters,ngrid=ngrid,**kwargs)
    
    def make_bounds(self,parameters,array=True,**kwargs):
        " Make the boundary conditions of the hyperparameters. "
        return self.bounds.get_bounds(parameters=parameters,array=array,**kwargs)
    
    def sample_thetas(self,parameters,npoints,**kwargs):
        " Draw random hyperparameter samples from the boundary conditions. "
        return self.bounds.sample_thetas(parameters=parameters,npoints=npoints,**kwargs)
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(local_optimizer=self.local_optimizer,
                        bounds=self.bounds,
                        maxiter=self.maxiter)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs
    

class RandomSamplingOptimizer(GlobalOptimizer):
    def __init__(self,local_optimizer=None,bounds=None,maxiter=5000,npoints=40,parallel=False,**kwargs):
        """
        The random sampling optimizer used for optimzing the objective function wrt. the hyperparameters.
        The random sampling optimizer samples the hyperparameters randomly from the boundary conditions
        and optimize all samples with the local optimizer.
        Parameters:
            local_optimizer : Local optimizer class
                A local optimization method.
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
            maxiter : int
                The maximum number of evaluations or iterations the global optimizer can use.
            npoints : int
                The number of hyperparameter points samled from the boundary conditions. 
            parallel : bool
                Whether to calculate the grid points in parallel over multiple CPUs.
        """
        # The gradient of the objective function is not used by the global optimizer
        self.jac=False
        # Set default bounds
        if bounds is None:
            from ..hpboundary.hptrans import VariableTransformation
            bounds=VariableTransformation(bounds=None)
        # Set default local optimizer
        if local_optimizer is None:
            from .localoptimizer import ScipyOptimizer
            local_optimizer=ScipyOptimizer(maxiter=int(maxiter/npoints),bounds=bounds,use_bounds=False)
        # Set all the arguments
        self.update_arguments(local_optimizer=local_optimizer,
                              bounds=bounds,
                              maxiter=maxiter,
                              npoints=npoints,
                              parallel=parallel,
                              **kwargs)

    def run(self,func,theta,parameters,model,X,Y,pdis,**kwargs):
        # Draw random hyperparameter samples
        thetas=np.array([theta])
        if self.npoints>1:
            thetas=self.sample_thetas(parameters,npoints=int(self.npoints-1))
            thetas=np.append(thetas,thetas,axis=0)
        # Make empty solution and lists
        sol=self.get_empty_solution()
        # Perform the local optimization for random samples
        sol=self.optimize_samples(sol,func,thetas,parameters,model,X,Y,pdis,**kwargs)
        return sol
    
    def update_arguments(self,local_optimizer=None,bounds=None,maxiter=None,npoints=None,parallel=None,**kwargs):
        """
        Update the optimizer with its arguments. The existing arguments are used if they are not given.
        Parameters:
            local_optimizer : Local optimizer class
                A local optimization method.
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            npoints : int
                The number of hyperparameter points samled from the boundary conditions. 
            parallel : bool
                Whether to calculate the grid points in parallel over multiple CPUs.
        Returns:
            self: The updated object itself.
        """
        if local_optimizer is not None:
            self.local_optimizer=local_optimizer.copy()
        if bounds is not None:
            self.bounds=bounds.copy()
            # Use the same boundary conditions in the local optimizer
            self.local_optimizer.update_arguments(bounds=self.bounds)
        if maxiter is not None:
            self.maxiter=int(maxiter)
        if parallel is not None:
            self.parallel=parallel
        if npoints is not None:
            if self.parallel:
                from ase.parallel import world
                # Make sure that the optimal number of points is used for the CPUs
                self.npoints=int(int(npoints/world.size)*world.size)
                if self.npoints==0:
                    self.npoints=world.size
            else:
                self.npoints=int(npoints)
        return self
    
    def optimize_samples(self,sol,func,thetas,parameters,model,X,Y,pdis,**kwargs):
        " Perform the local optimization of the random samples. "
        # Check if the optimization should be performed in parallel
        if self.parallel:
            return self.optimize_samples_parallel(sol,func,thetas,parameters,model,X,Y,pdis,**kwargs)
        for theta in thetas:
            # Check if the maximum number of iterations is used
            if sol['nfev']>=self.maxiter:
                break
            # Do local optimization
            sol_s=self.run_local_opt(func,theta,parameters,model,X,Y,pdis,**kwargs)
            # Update the solution if it is better
            sol=self.compare_solutions(sol,sol_s)
        # Update the total number of iterations
        sol['nit']=len(thetas)
        # Get the all best time best solution 
        return self.get_final_solution(sol,func,parameters,model,X,Y,pdis)
    
    def optimize_samples_parallel(self,sol,func,thetas,parameters,model,X,Y,pdis,**kwargs):
        " Perform the local optimization of the random samples in parallel. "
        from ase.parallel import world
        rank,size=world.rank,world.size
        for t,theta in enumerate(thetas):
            if rank==t%size:
                # Check if the maximum number of iterations is used
                if sol['nfev']>=self.maxiter:
                    break
                # Do local optimization in parallel
                sol_s=self.run_local_opt(func,theta,parameters,model,X,Y,pdis,**kwargs)
                # Update the solution if it is better
                sol=self.compare_solutions(sol,sol_s)
        # Update the total number of iterations
        sol['nit']=len(thetas)
        # Get the all best time best solution for all CPUs and broadcast it
        return self.get_final_solution(sol,func,parameters,model,X,Y,pdis)
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(local_optimizer=self.local_optimizer,
                        bounds=self.bounds,
                        maxiter=self.maxiter,
                        npoints=self.npoints)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs


class GridOptimizer(GlobalOptimizer):
    def __init__(self,local_optimizer=None,bounds=None,maxiter=5000,n_each_dim=None,optimize=True,parallel=False,**kwargs):
        """
        The grid optimizer used for optimzing the objective function wrt. the hyperparameters.
        The grid optimizer makes a grid in the hyperparameter space from the boundary conditions and evaluate them.
        The grid point with the lowest function value can be optimized with the local optimizer.
        Parameters:
            local_optimizer : Local optimizer class
                A local optimization method.
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
            maxiter : int
                The maximum number of evaluations or iterations the global optimizer can use.
            n_each_dim : int or (H) list
                An integer or a list with number of grid points in each dimension of the hyperparameters.
            optimize : bool
                Whether to perform a local optimization on the best found solution. 
            parallel : bool
                Whether to calculate the grid points in parallel over multiple CPUs.
        """
        # The gradient of the objective function is not used by the global optimizer
        self.jac=False
        # Set default bounds
        if bounds is None:
            from ..hpboundary.hptrans import VariableTransformation
            bounds=VariableTransformation(bounds=None)
        # Set default local optimizer
        if local_optimizer is None:
            from .localoptimizer import ScipyOptimizer
            local_optimizer=ScipyOptimizer(maxiter=maxiter,bounds=bounds,use_bounds=False)
        # Set n_each_dim as default
        self.n_each_dim=None
        # Set all the arguments
        self.update_arguments(local_optimizer=local_optimizer,
                              bounds=bounds,
                              maxiter=maxiter,
                              n_each_dim=n_each_dim,
                              optimize=optimize,
                              parallel=parallel,
                              **kwargs)

    def run(self,func,theta,parameters,model,X,Y,pdis,**kwargs):
        # Number of points per dimension
        n_each_dim=self.get_n_each_dim(len(theta))
        # Make grid either with the same or different numbers in each dimension
        lines=self.make_lines(parameters,ngrid=n_each_dim)
        thetas=np.append([theta],self.make_grid(lines,maxiter=int(self.maxiter-1)),axis=0)
        # Check if the number of points is well parallized if it is used
        thetas=self.check_npoints(thetas)
        # Make empty solution and lists
        sol=self.get_empty_solution()
        # Get the function arguments
        func_args=self.get_func_arguments(parameters,model,X,Y,pdis,jac=False,**kwargs)
        # Calculate the grid points
        f_list=self.calculate_values(thetas,func,func_args=func_args)
        # Find the minimum function value
        sol=self.get_minimum(sol,thetas,f_list)
        sol=self.get_final_solution(sol,func,parameters,model,X,Y,pdis)
        # Perform the local optimization for the minimum function value
        sol=self.optimize_minimum(sol,func,parameters,model,X,Y,pdis,**kwargs)
        return sol
    
    def update_arguments(self,local_optimizer=None,bounds=None,maxiter=None,n_each_dim=None,optimize=None,parallel=None,**kwargs):
        """
        Update the optimizer with its arguments. The existing arguments are used if they are not given.
        Parameters:
            local_optimizer : Local optimizer class
                A local optimization method.
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
            maxiter : int
                The maximum number of evaluations or iterations the global optimizer can use.
            n_each_dim : int or (H) list
                An integer or a list with number of grid points in each dimension of the hyperparameters.
            optimize : bool
                Whether to perform a local optimization on the best found solution. 
            parallel : bool
                Whether to calculate the grid points in parallel over multiple CPUs.
        Returns:
            self: The updated object itself.
        """
        if local_optimizer is not None:
            self.local_optimizer=local_optimizer.copy()
        if bounds is not None:
            self.bounds=bounds.copy()
            # Use the same boundary conditions in the local optimizer
            self.local_optimizer.update_arguments(bounds=self.bounds)
        if maxiter is not None:
            self.maxiter=int(maxiter)
        if parallel is not None:
            self.parallel=parallel
        if n_each_dim is not None:
            if isinstance(n_each_dim,(list,np.ndarray)):
                self.n_each_dim=n_each_dim.copy()
            else:
                self.n_each_dim=n_each_dim
        if optimize is not None:
            self.optimize=optimize
        return self
    
    def make_grid(self,lines,maxiter=5000):
        " Make a grid in multi-dimensions from a list of 1D grids in each dimension. "
        lines=np.array(lines)
        if len(lines.shape)<2:
            lines=lines.reshape(1,-1)
        # Number of combinations
        combi=1
        for i in [len(line) for line in lines]:
            combi*=i
        if combi<maxiter:
            maxiter=combi
        # If there is a low probability to find grid points randomly the entire grid are calculated
        if (1-(maxiter/combi))<0.99:
            X=lines[0].reshape(-1,1)
            lines=lines[1:]
            for line in lines:
                dim_X=len(X)
                X=np.concatenate([X]*len(line),axis=0)
                X=np.concatenate([X,np.sort(np.concatenate([line.reshape(-1)]*dim_X,axis=0)).reshape(-1,1)],axis=1)
            return np.random.permutation(X)[:maxiter]
        # Randomly sample the grid points
        X=np.array([np.random.choice(line,size=maxiter) for line in lines]).T
        X=np.unique(X,axis=0)
        while len(X)<maxiter:
            x=np.array([np.random.choice(line,size=1) for line in lines]).T
            X=np.append(X,x,axis=0)
            X=np.unique(X,axis=0)
        return X[:maxiter]
    
    def optimize_minimum(self,sol,func,parameters,model,X,Y,pdis,**kwargs):
        " Perform the local optimization of the found minimum. "
        # Check if optimization should be used
        if not self.optimize:
            return sol
        # Check if all iterations have been used
        if sol['nfev']>=self.maxiter:
            return sol
        # Perform local optimization
        sol_s=self.run_local_opt(func,sol['x'],parameters,model,X,Y,pdis,**kwargs)
        # Update the solution if it is better
        sol=self.compare_solutions(sol,sol_s)
        # Update the number of used iterations
        sol['nit']+=1
        return self.get_final_solution(sol,func,parameters,model,X,Y,pdis)
    
    def get_n_each_dim(self,dim,**kwargs):
        " Number of points per dimension. "
        if self.n_each_dim is None:
            n_each_dim=int((self.maxiter-1)**(1/dim))
            n_each_dim=n_each_dim if n_each_dim>1 else 1
        else:
            n_each_dim=self.n_each_dim
        return n_each_dim
    
    def check_npoints(self,thetas,**kwargs):
        " Check if the number of points is well parallized if it is used. "
        if self.parallel:
            from ase.parallel import world
            npoints=int(int(len(thetas)/world.size)*world.size)
            if npoints==0:
                npoints=world.size
            return thetas[:npoints]
        return thetas
    
    def get_minimum(self,sol,thetas,f_list,**kwargs):
        " Find the minimum function value and update the solution. "
        # Find the minimum function value
        i_min=np.nanargmin(f_list)
        # Get the number of used iterations
        thetas_len=len(thetas)
        # Update the number of used iterations
        sol['nfev']+=thetas_len
        sol['nit']+=thetas_len
        # Check if a better point is found
        if f_list[i_min]>sol['fun']:
            return sol
        # Update the solution if a better point is found
        sol['fun']=f_list[i_min]
        sol['x']=thetas[i_min].copy()
        sol['message']="Lower function value found."
        return sol
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(local_optimizer=self.local_optimizer,
                        bounds=self.bounds,
                        maxiter=self.maxiter,
                        n_each_dim=self.n_each_dim,
                        optimize=self.optimize,
                        parallel=self.parallel)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs


class IterativeLineOptimizer(GridOptimizer):
    def __init__(self,local_optimizer=None,bounds=None,maxiter=5000,n_each_dim=None,loops=3,calculate_init=False,optimize=True,parallel=False,**kwargs):
        """
        The iteratively line optimizer used for optimzing the objective function wrt. the hyperparameters.
        The iteratively line optimizer makes a 1D grid in each dimension of the hyperparameter space from the boundary conditions.
        The grid points are then evaluated and the best value updates the hyperparameter in the specific dimension.
        This process is done iteratively over all dimensions and in loops. 
        The grid point with the lowest function value can be optimized with the local optimizer.
        Parameters:
            local_optimizer : Local optimizer class
                A local optimization method.
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
            maxiter : int
                The maximum number of evaluations or iterations the global optimizer can use.
            n_each_dim : int or (H) list
                An integer or a list with number of grid points in each dimension of the hyperparameters.
            loops : int
                The number of times all the hyperparameter dimensions have been searched.
            calculate_init : bool
                Whether to calculate the initial given hyperparameters.
                If it is parallelized, all CPUs will calculate this point. 
            optimize : bool
                Whether to perform a local optimization on the best found solution. 
            parallel : bool
                Whether to calculate the grid points in parallel over multiple CPUs.
        """
        super().__init__(local_optimizer=local_optimizer,
                         bounds=bounds,
                         maxiter=maxiter,
                         n_each_dim=n_each_dim,
                         loops=loops,
                         calculate_init=calculate_init,
                         optimize=optimize,
                         parallel=parallel,
                         **kwargs)

    def run(self,func,theta,parameters,model,X,Y,pdis,**kwargs):
        # Number of points per dimension
        n_each_dim=self.get_n_each_dim(len(theta))
        # Make grid either with the same or different numbers in each dimension
        lines=self.make_lines(parameters,ngrid=n_each_dim)
        # Get the function arguments
        func_args=self.get_func_arguments(parameters,model,X,Y,pdis,jac=False,**kwargs)
        # Calculate the grid points in the iterative grid/line search
        sol=self.iterative_line(theta,lines,func,func_args=func_args)
        sol=self.get_final_solution(sol,func,parameters,model,X,Y,pdis)
        # Perform the local optimization for the minimum function value
        sol=self.optimize_minimum(sol,func,parameters,model,X,Y,pdis,**kwargs)
        return sol
    
    def update_arguments(self,local_optimizer=None,bounds=None,maxiter=None,n_each_dim=None,loops=None,calculate_init=None,optimize=None,parallel=None,**kwargs):
        """
        Update the optimizer with its arguments. The existing arguments are used if they are not given.
        Parameters:
            local_optimizer : Local optimizer class
                A local optimization method.
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
            maxiter : int
                The maximum number of evaluations or iterations the global optimizer can use.
            n_each_dim : int or (H) list
                An integer or a list with number of grid points in each dimension of the hyperparameters.
            loops : int
                The number of times all the hyperparameter dimensions have been searched.
            calculate_init : bool
                Whether to calculate the initial given hyperparameters.
                If it is parallelized, all CPUs will calculate this point. 
            optimize : bool
                Whether to perform a local optimization on the best found solution. 
            parallel : bool
                Whether to calculate the grid points in parallel over multiple CPUs.
        Returns:
            self: The updated object itself.
        """
        if local_optimizer is not None:
            self.local_optimizer=local_optimizer.copy()
        if bounds is not None:
            self.bounds=bounds.copy()
            # Use the same boundary conditions in the local optimizer
            self.local_optimizer.update_arguments(bounds=self.bounds)
        if maxiter is not None:
            self.maxiter=int(maxiter)
        if parallel is not None:
            self.parallel=parallel
        if loops is not None:
            self.loops=int(loops)
        if calculate_init is not None:
            self.calculate_init=calculate_init
        if n_each_dim is not None:
            if isinstance(n_each_dim,(list,np.ndarray)):
                if np.sum(n_each_dim)*self.loops>self.maxiter:
                    self.n_each_dim=self.get_n_each_dim(len(n_each_dim))
                else:
                    self.n_each_dim=n_each_dim.copy()
            else:
                self.n_each_dim=n_each_dim
        if optimize is not None:
            self.optimize=optimize
        return self
    
    def iterative_line(self,theta,lines,func,func_args=(),**kwargs):
        " Perform iteratively grid/line search. "
        # Make an initial solution
        if self.calculate_init:
            sol=self.get_initial_solution(theta,func,func_args=func_args)
        else:
            sol=self.get_empty_solution()
        # Get the dimension list
        dims=list(range(len(lines)))
        # Set initidal dimension
        d=None
        # Perform loops
        for l in range(self.loops):
            # Permute the dimensions
            dim_perm=np.random.permutation(dims)
            # Make sure the same dimension is not used after each other
            if dim_perm[0]==d:
                dim_perm=dim_perm[1:]
            for d in dim_perm:
                # Make the hyperparameter changes to the specific dimension
                thetas=np.tile(theta,(len(lines[d]),1))
                thetas[:,d]=lines[d].copy()
                f_list=self.calculate_values(thetas,func,func_args=func_args)
                sol=self.get_minimum(sol,thetas,f_list)
                theta=sol['x'].copy()
        return sol
    
    def get_n_each_dim(self,dim):
        " Number of points per dimension. "
        if self.n_each_dim is None:
            n_each_dim=int(self.maxiter/(self.loops*dim))
            n_each_dim=n_each_dim if n_each_dim>1 else 1
        else:
            n_each_dim=self.n_each_dim
        if self.parallel:
            return self.get_n_each_dim_parallel(n_each_dim)
        return n_each_dim
    
    def get_n_each_dim_parallel(self,n_each_dim):
        " Number of points per dimension if it is parallelized. "
        from ase.parallel import world
        if isinstance(n_each_dim,(list,np.ndarray)):
            for d,n_dim in enumerate(n_each_dim):
                n_each_dim[d]=int(int(n_dim/world.size)*world.size)
                if n_each_dim[d]==0:
                    n_each_dim[d]=world.size
        else:
            n_each_dim=int(int(n_each_dim/world.size)*world.size)
            if n_each_dim==0:
                n_each_dim=world.size
        return n_each_dim
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(local_optimizer=self.local_optimizer,
                        bounds=self.bounds,
                        maxiter=self.maxiter,
                        n_each_dim=self.n_each_dim,
                        loops=self.loops,
                        calculate_init=self.calculate_init,
                        optimize=self.optimize,
                        parallel=self.parallel)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs


class BasinOptimizer(GlobalOptimizer):
    def __init__(self,maxiter=5000,jac=True,opt_kwargs={},local_kwargs={},**kwargs):
        """
        The basin-hopping optimizer used for optimzing the objective function wrt. the hyperparameters.
        The basin-hopping optimizer is a wrapper to SciPy's basinhopping.
        (https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.basinhopping.html)
        No local optimizer and boundary conditions are given to this optimizer. 
        The local optimizer is set by keywords in the local_kwargs and it uses SciPy's minimizer.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the global optimizer can use.
            jac : bool
                Whether to use the gradient of the objective function wrt. the hyperparameters.
            opt_kwargs : dict
                A dictionary with the arguments and keywords given to SciPy's basinhopping.
            local_kwargs : dict
                A dictionary with the arguments and keywords given to SciPy's local minimizer.
        """
        # This global optimizer can not be parallelized
        self.parallel=False
        # Set default arguments for SciPy's basinhopping
        self.opt_kwargs=dict(niter=5,interval=10,T=1.0,stepsize=0.1,niter_success=None)
        # Set default arguments for SciPy's local minimizer
        self.local_kwargs=dict(options={'maxiter':int(maxiter/self.opt_kwargs['niter'])})
        # Set all the arguments
        self.update_arguments(maxiter=maxiter,
                              jac=jac,
                              opt_kwargs=opt_kwargs,
                              local_kwargs=local_kwargs,
                              **kwargs)

    def run(self,func,theta,parameters,model,X,Y,pdis,**kwargs):
        from scipy.optimize import basinhopping
        # Get the function arguments
        func_args=self.get_func_arguments(parameters,model,X,Y,pdis,self.jac,**kwargs)
        # Get the function that evaluate the objective function
        fun=self.get_fun(func)
        # Set the minimizer kwargs
        minimizer_kwargs=dict(args=func_args,jac=self.jac,**self.local_kwargs)
        # Do the basin-hopping
        sol=basinhopping(fun,x0=theta,minimizer_kwargs=minimizer_kwargs,**self.opt_kwargs)
        return self.get_final_solution(sol,func,parameters,model,X,Y,pdis)
    
    def update_arguments(self,maxiter=None,jac=None,opt_kwargs=None,local_kwargs=None,**kwargs):
        """
        Update the optimizer with its arguments. The existing arguments are used if they are not given.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the global optimizer can use.
            jac : bool
                Whether to use the gradient of the objective function wrt. the hyperparameters.
            opt_kwargs : dict
                A dictionary with the arguments and keywords given to SciPy's basinhopping.
            local_kwargs : dict
                A dictionary with the arguments and keywords given to SciPy's local minimizer.
        Returns:
            self: The updated object itself.
        """
        if maxiter is not None:
            self.maxiter=int(maxiter)
        if jac is not None:
            self.jac=jac
        if opt_kwargs is not None:
            self.opt_kwargs.update(opt_kwargs)
        if local_kwargs is not None:
            if 'options' in local_kwargs:
                local_no_options={key:value for key,value in local_kwargs.items() if key!='options'}
                self.local_kwargs.update(local_no_options)
                self.local_kwargs['options'].update(local_kwargs['options'])
            else:
                self.local_kwargs.update(local_kwargs)
        # Make sure not to many iterations are used in average
        maxiter_niter=int(self.maxiter/self.opt_kwargs['niter'])
        if maxiter_niter<self.local_kwargs['options']['maxiter']:
            self.local_kwargs['options']['maxiter']=maxiter_niter
        return self
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(maxiter=self.maxiter,jac=self.jac,
                        opt_kwargs=self.opt_kwargs,
                        local_kwargs=self.local_kwargs)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs
    

class AnneallingOptimizer(GlobalOptimizer):
    def __init__(self,bounds=None,maxiter=5000,jac=True,opt_kwargs={},local_kwargs={},**kwargs):
        """
        The simulated annealing optimizer used for optimzing the objective function wrt. the hyperparameters.
        The simulated annealing optimizer is a wrapper to SciPy's dual_annealing.
        (https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.dual_annealing.html)
        No local optimizer is given to this optimizer. 
        The local optimizer is set by keywords in the local_kwargs and it uses SciPy's minimizer.
        Parameters:
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
            maxiter : int
                The maximum number of evaluations or iterations the global optimizer can use.
            jac : bool
                Whether to use the gradient of the objective function wrt. the hyperparameters.
            opt_kwargs : dict
                A dictionary with the arguments and keywords given to SciPy's dual_annealing.
            local_kwargs : dict
                A dictionary with the arguments and keywords given to SciPy's local minimizer.
        """
        # Set default bounds
        if bounds is None:
            from ..hpboundary.educated import EducatedBoundaries
            bounds=EducatedBoundaries(log=True)
        # This global optimizer can not be parallelized
        self.parallel=False
        # Set default arguments for SciPy's dual_annealing
        self.opt_kwargs=dict(initial_temp=5230.0,restart_temp_ratio=2e-05,
                             visit=2.62,accept=-5.0,seed=None,no_local_search=False)
        # Set default arguments for SciPy's local minimizer
        self.local_kwargs=dict(options={})
        # Set all the arguments
        self.update_arguments(bounds=bounds,
                              maxiter=maxiter,
                              jac=jac,
                              opt_kwargs=opt_kwargs,
                              local_kwargs=local_kwargs,
                              **kwargs)

    def run(self,func,theta,parameters,model,X,Y,pdis,**kwargs):
        from scipy.optimize import dual_annealing
        # Get the function arguments
        func_args=self.get_func_arguments(parameters,model,X,Y,pdis,jac=False,**kwargs)
        # Get the function
        fun=self.get_fun(func)
        # Set the minimizer kwargs
        minimizer_kwargs=dict(jac=False,**self.local_kwargs)
        # Make boundary conditions
        bounds=self.make_bounds(parameters,array=True)
        # Do the dual simulated annealing
        sol=dual_annealing(fun,bounds=bounds,
                           x0=theta,args=func_args,
                           maxiter=self.maxiter,maxfun=self.maxiter,
                           minimizer_kwargs=minimizer_kwargs,**self.opt_kwargs)        
        return self.get_final_solution(sol,func,parameters,model,X,Y,pdis)
    
    def update_arguments(self,bounds=None,maxiter=None,jac=None,opt_kwargs=None,local_kwargs=None,**kwargs):
        """
        Update the optimizer with its arguments. The existing arguments are used if they are not given.
        Parameters:
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
            maxiter : int
                The maximum number of evaluations or iterations the global optimizer can use.
            jac : bool
                Whether to use the gradient of the objective function wrt. the hyperparameters.
            opt_kwargs : dict
                A dictionary with the arguments and keywords given to SciPy's dual_annealing.
            local_kwargs : dict
                A dictionary with the arguments and keywords given to SciPy's local minimizer.
        Returns:
            self: The updated object itself.
        """
        if bounds is not None:
            self.bounds=bounds.copy()
        if maxiter is not None:
            self.maxiter=int(maxiter)
        if jac is not None:
            self.jac=jac
        if opt_kwargs is not None:
            self.opt_kwargs.update(opt_kwargs)
        if local_kwargs is not None:
            if 'options' in local_kwargs:
                local_no_options={key:value for key,value in local_kwargs.items() if key!='options'}
                self.local_kwargs.update(local_no_options)
                self.local_kwargs['options'].update(local_kwargs['options'])
            else:
                self.local_kwargs.update(local_kwargs)
        return self
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(bounds=self.bounds,maxiter=self.maxiter,
                        jac=self.jac,opt_kwargs=self.opt_kwargs,
                        local_kwargs=self.local_kwargs)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs


class AnneallingTransOptimizer(AnneallingOptimizer):
    def __init__(self,bounds=None,maxiter=5000,jac=True,opt_kwargs={},local_kwargs={},**kwargs):
        """
        The simulated annealing optimizer used for optimzing the objective function wrt. the hyperparameters.
        The simulated annealing optimizer is a wrapper to SciPy's dual_annealing.
        (https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.dual_annealing.html)
        No local optimizer is given to this optimizer. 
        The local optimizer is set by keywords in the local_kwargs and it uses SciPy's minimizer.
        This simulated annealing optimizer uses variable transformation of the hyperparameters to search the space.
        Parameters:
            bounds : VariableTransformation class
                A class of the variable transformation of the hyperparameters.
            maxiter : int
                The maximum number of evaluations or iterations the global optimizer can use.
            jac : bool
                Whether to use the gradient of the objective function wrt. the hyperparameters.
            opt_kwargs : dict
                A dictionary with the arguments and keywords given to SciPy's dual_annealing.
            local_kwargs : dict
                A dictionary with the arguments and keywords given to SciPy's local minimizer.
        """
        # Set default bounds
        if bounds is None:
            from ..hpboundary.hptrans import VariableTransformation
            bounds=VariableTransformation(bounds=None)
        # This global optimizer can not be parallelized
        self.parallel=False
        # Set default arguments for SciPy's dual_annealing
        self.opt_kwargs=dict(initial_temp=5230.0,restart_temp_ratio=2e-05,
                             visit=2.62,accept=-5.0,seed=None,no_local_search=False)
        # Set default arguments for SciPy's local minimizer
        self.local_kwargs=dict(options={})
        # Set all the arguments
        self.update_arguments(bounds=bounds,
                              maxiter=maxiter,
                              jac=jac,
                              opt_kwargs=opt_kwargs,
                              local_kwargs=local_kwargs,
                              **kwargs)

    def run(self,func,theta,parameters,model,X,Y,pdis,**kwargs):
        from scipy.optimize import dual_annealing
        # Get the function arguments for the wrappers
        func_args_w=self.get_wrapper_arguments(func,parameters,model,X,Y,pdis,jac=self.jac,**kwargs)
        # Set the minimizer kwargs
        minimizer_kwargs=dict(jac=False,**self.local_kwargs)
        # Make boundary conditions
        bounds=self.make_bounds(parameters,array=True,transformed=True)
        # Do the dual simulated annealing
        sol=dual_annealing(self.func_vartrans,bounds=bounds,
                           x0=theta,args=func_args_w,
                           maxiter=self.maxiter,maxfun=self.maxiter,
                           minimizer_kwargs=minimizer_kwargs,**self.opt_kwargs)
        sol=self.get_final_solution(sol,func,parameters,model,X,Y,pdis)
        # Retransform hyperparameters
        sol=self.transform_solution(sol)
        return sol
    
    def update_arguments(self,bounds=None,maxiter=None,jac=None,opt_kwargs=None,local_kwargs=None,**kwargs):
        """
        Update the optimizer with its arguments. The existing arguments are used if they are not given.
        Parameters:
            bounds : VariableTransformation class
                A class of the variable transformation of the hyperparameters.
            maxiter : int
                The maximum number of evaluations or iterations the global optimizer can use.
            jac : bool
                Whether to use the gradient of the objective function wrt. the hyperparameters.
            opt_kwargs : dict
                A dictionary with the arguments and keywords given to SciPy's dual_annealing.
            local_kwargs : dict
                A dictionary with the arguments and keywords given to SciPy's local minimizer.
        Returns:
            self: The updated object itself.
        """
        if bounds is not None:
            from ..hpboundary.hptrans import VariableTransformation
            if not isinstance(bounds,VariableTransformation):
                raise Exception('A variable transformation as bounds has to be used!')
            self.bounds=bounds.copy()
        if maxiter is not None:
            self.maxiter=int(maxiter)
        if jac is not None:
            self.jac=jac
        if opt_kwargs is not None:
            self.opt_kwargs.update(opt_kwargs)
        if local_kwargs is not None:
            if 'options' in local_kwargs:
                local_no_options={key:value for key,value in local_kwargs.items() if key!='options'}
                self.local_kwargs.update(local_no_options)
                self.local_kwargs['options'].update(local_kwargs['options'])
            else:
                self.local_kwargs.update(local_kwargs)
        return self

    def func_vartrans(self,ti,fun,parameters,func_args=(),**kwargs):
        " Objective function called for simulated annealing, where hyperparameters are transformed. "
        theta=self.reverse_trasformation(ti,parameters)
        return fun(theta,*func_args)
    
    def reverse_trasformation(self,ti,parameters,**kwargs):
        " Transform the variable transformed hyperparameters back to hyperparameter log-space. "
        ti=np.where(ti<1.0,np.where(ti>0.0,ti,self.bounds.eps),1.00-self.bounds.eps)
        t=self.make_hp(ti,parameters)
        theta=self.bounds.reverse_trasformation(t,array=True)
        return theta
    
    def transform_solution(self,sol,**kwargs):
        " Retransform the variable transformed hyperparameters in the solution back to hyperparameter log-space. "
        sol['x']=self.bounds.reverse_trasformation(sol['hp'],array=True)
        sol['hp']=self.bounds.reverse_trasformation(sol['hp'],array=False)
        return sol
    
    def get_wrapper_arguments(self,func,parameters,model,X,Y,pdis,jac,**kwargs):
        " Get the function arguments for the wrappers. "
        # Get the function arguments
        func_args=self.get_func_arguments(parameters,model,X,Y,pdis,jac=False,**kwargs)
        # Get the function that evaluate the objective function
        fun=self.get_fun(func)
        # Get the function arguments for the wrappers
        func_args_w=(fun,parameters,func_args)
        return func_args_w


class FactorizedOptimizer(GlobalOptimizer):
    def __init__(self,line_optimizer=None,bounds=None,maxiter=5000,ngrid=80,calculate_init=False,parallel=False,**kwargs):
        """
        The factorized optimizer used for optimzing the objective function wrt. the hyperparameters.
        The factorized optimizer makes a 1D grid for each hyperparameter from the boundary conditions. 
        The hyperparameters are then optimized with a line search optimizer. 
        The line search optimizer optimizes only one of the hyperparameters and 
        it therefore relies on a factorization method as the objective function. 
        Parameters:
            line_optimizer : Line search optimizer class
                A line search optimization method.
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
            maxiter : int
                The maximum number of evaluations or iterations the global optimizer can use.
            ngrid : int
                The number of grid points of the hyperparameter that is optimized.
            calculate_init : bool
                Whether to calculate the initial given hyperparameters.
                If it is parallelized, all CPUs will calculate this point. 
            parallel : bool
                Whether to calculate the grid points in parallel over multiple CPUs.
        """
        # The gradient of the objective function is not used by the global optimizer
        self.jac=False
        # Set default bounds
        if bounds is None:
            from ..hpboundary.hptrans import VariableTransformation
            bounds=VariableTransformation(bounds=None)
        # Set default line optimizer
        if line_optimizer is None:
            from .linesearcher import GoldenSearch
            line_optimizer=GoldenSearch(maxiter=int(maxiter),parallel=parallel)
        # Set all the arguments
        self.update_arguments(line_optimizer=line_optimizer,
                              bounds=bounds,
                              maxiter=maxiter,
                              ngrid=ngrid,
                              calculate_init=calculate_init,
                              parallel=parallel,
                              **kwargs)

    def run(self,func,theta,parameters,model,X,Y,pdis,**kwargs):
        # Make an initial solution or use an empty solution
        if self.calculate_init:
            func_args=self.get_func_arguments(parameters,model,X,Y,pdis,jac=False,**kwargs)
            sol=self.get_initial_solution(theta,func,func_args=func_args)
        else:
            sol=self.get_empty_solution()
        # Make the lines of the hyperparameters
        lines=np.array(self.make_lines(parameters,ngrid=self.ngrid)).T
        # Optimize the hyperparameters with the line search
        sol_s=self.run_line_opt(func,lines,parameters,model,X,Y,pdis,**kwargs)
        # Update the solution if it is better
        sol=self.compare_solutions(sol,sol_s)
        # Change the solution message
        if sol['success']:
            sol['message']='Local optimization is converged.'
        else:
            sol['message']='Local optimization is not converged.'
        return self.get_final_solution(sol,func,parameters,model,X,Y,pdis)
    
    def update_arguments(self,line_optimizer=None,bounds=None,maxiter=None,ngrid=None,calculate_init=None,parallel=None,**kwargs):
        """
        Update the optimizer with its arguments. The existing arguments are used if they are not given.
        Parameters:
            line_optimizer : Line search optimizer class
                A line search optimization method.
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
            maxiter : int
                The maximum number of evaluations or iterations the global optimizer can use.
            ngrid : int
                The number of grid points of the hyperparameter that is optimized.
            calculate_init : bool
                Whether to calculate the initial given hyperparameters.
                If it is parallelized, all CPUs will calculate this point. 
            parallel : bool
                Whether to calculate the grid points in parallel over multiple CPUs.
        Returns:
            self: The updated object itself.
        """
        if line_optimizer is not None:
            self.line_optimizer=line_optimizer.copy()
        if bounds is not None:
            self.bounds=bounds.copy()
        if maxiter is not None:
            self.maxiter=int(maxiter)
        if parallel is not None:
            self.parallel=parallel
        if ngrid is not None:
            if self.parallel:
                from ase.parallel import world
                self.ngrid=int(int(ngrid/world.size)*world.size)
                if self.ngrid==0:
                    self.ngrid=world.size
            else:
                self.ngrid=ngrid
        if calculate_init is not None:
            self.calculate_init=calculate_init
        return self
    
    def run_line_opt(self,func,lines,parameters,model,X,Y,pdis,**kwargs):
        " Run the line search optimization. "
        return self.line_optimizer.run(func,lines,parameters,model,X,Y,pdis,**kwargs)
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(line_optimizer=self.line_optimizer,
                        bounds=self.bounds,
                        maxiter=self.maxiter,
                        ngrid=self.ngrid,
                        calculate_init=self.calculate_init,
                        parallel=self.parallel)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs




import numpy as np
from scipy.linalg import cho_factor,cho_solve
from numpy.linalg import eigh

class ObjectiveFuction:
    def __init__(self,get_prior_mean=False,**kwargs):
        """ 
        The objective function that is used to optimize the hyperparameters. 
        Parameters:
            get_prior_mean : bool
                Whether to get the parameters of the prior mean in the solution.
        """
        # Set descriptor of the objective function
        self.use_analytic_prefactor=False
        self.use_optimized_noise=False
        # Set the arguments
        self.update_arguments(get_prior_mean=get_prior_mean,**kwargs)

    def function(self,theta,parameters,model,X,Y,pdis=None,jac=False,**kwargs):
        """ 
        The function call that calculate the objective function. 
        Parameters:
            theta: (H) array of floats
                An array with the hyperparameter values used for the objective function.
            parameters: (H) list of strings
                A list of names of the hyperparameters.
            model: Model
                The Machine Learning Model with kernel and prior that are optimized.
            X: (N,D) array
                Training features with N data points and D dimensions.
            Y: (N,1) array or (N,D+1) array
                Training targets without or with derivatives with N data points.
            pdis: dict
                A dict of prior distributions for each hyperparameter type.
            jac: bool
                Whether to get the derivatives of the objective function wrt. the hyperparameters. 
        Returns:
            float: The objective function value.
            and/or
            (H) array: The derivative of the objective function value wrt. the hyperparameters if jac=True.
        """
        raise NotImplementedError()
    
    def derivative(self,**kwargs):
        """ 
        The derivative of the objective function wrt. the hyperparameters. 
        """
        raise NotImplementedError()
    
    def update_arguments(self,get_prior_mean=None,**kwargs):
        """
        Update the objective function with its arguments. The existing arguments are used if they are not given.
        Parameters:
            get_prior_mean : bool
                Whether to get the parameters of the prior mean in the solution.
        Returns:
            self: The updated object itself.
        """
        if get_prior_mean is not None:
            self.get_prior_mean=get_prior_mean
        # Always reset the solution when the objective function is changed 
        self.reset_solution()
        return self
    
    def reset_solution(self):
        """ 
        Reset the solution of the optimization in terms of the hyperparameters and model. 
        """
        self.sol={'fun':np.inf,'x':np.array([]),'hp':{}}
        return self
    
    def update_solution(self,fun,theta,hp,model,jac=False,deriv=None,**kwargs):
        """
        Update the solution of the optimization in terms of hyperparameters and model.
        The lowest objective function value is stored togeher with its hyperparameters.
        The prior mean can also be saved if get_prior_mean=True.
        """
        if fun<self.sol['fun']:
            self.sol['fun']=fun
            self.sol['x']=theta.copy()
            self.sol['hp']=hp.copy()
            if jac:
                self.sol['jac']=deriv.copy()
            if self.get_prior_mean:
                self.sol['prior']=self.get_prior_parameters(model)
        return self.sol
    
    def get_solution(self,sol,parameters,model,X,Y,pdis=None,**kwargs):
        """ 
        Get the solution of the optimization in terms of hyperparameters and model. 
        """
        if self.sol['fun']>sol['fun']:
            sol['hp']=self.make_hp(sol['x'],parameters)[0]
            if self.get_prior_mean:
                sol['prior']=self.get_prior_parameters(model,X=X,Y=Y)
            return sol
        sol['fun']=self.sol['fun']
        sol['x']=self.sol['x'].copy()
        sol['hp']=self.sol['hp'].copy()
        if 'jac' in self.sol.keys():
            sol['jac']=self.sol['jac'].copy()
        if 'prior' in self.sol.keys():
            sol['prior']=self.sol['prior'].copy()
        return sol
    
    def get_stored_solution(self,**kwargs):
        """
        Get the stored solution of the optimization of the hyperparameters within only the objective function.
        """
        return self.sol

    def make_hp(self,theta,parameters,**kwargs):
        " Make hyperparameter dictionary from lists"
        theta,parameters=np.array(theta),np.array(parameters)
        parameters_set=sorted(set(parameters))
        hp={para_s:self.numeric_limits(theta[parameters==para_s]) for para_s in parameters_set}
        return hp,parameters_set
    
    def get_hyperparams(self,model,**kwargs):
        " Get the hyperparameters for the model and the kernel. "
        return model.get_hyperparams()
    
    def numeric_limits(self,array,dh=0.1*np.log(np.finfo(float).max)):
        " Replace hyperparameters if they are outside of the numeric limits in log-space. "
        return np.where(-dh<array,np.where(array<dh,array,dh),-dh)
    
    def update_model(self,model,hp,**kwargs):
        " Update the the machine learning model with the hyperparameters. "
        model.set_hyperparams(hp)
        return model

    def kxx_reg(self,model,X,**kwargs):
        " Get covariance matrix with regularization. "
        KXX=model.kernel(X,get_derivatives=model.use_derivatives)
        KXX_n=model.add_regularization(KXX,len(X),overwrite=False)
        return KXX_n,KXX,len(KXX)
    
    def kxx_corr(self,model,X,**kwargs):
        " Get covariance matrix with or without noise correction. "
        # Calculate the kernel with and without noise
        KXX=model.kernel(X,get_derivatives=model.use_derivatives)
        n_data=len(KXX)
        KXX=self.add_correction(model,KXX,n_data)
        return KXX,n_data
    
    def add_correction(self,model,KXX,n_data,**kwargs):
        " Add noise correction to covariance matrix. "
        corr=model.get_correction(np.diag(KXX))
        if corr>0.0:
            KXX[range(n_data),range(n_data)]+=corr
        return KXX
        
    def y_prior(self,X,Y,model,L=None,low=None,**kwargs):
        " Update prior and subtract to target. "
        Y_p=Y.copy()
        model.prior.update(X,Y_p,L=L,low=low,**kwargs)
        if model.use_derivatives:
            return (Y_p-model.prior.get(X,Y_p,get_derivatives=True)).T.reshape(-1,1)
        return (Y_p-model.prior.get(X,Y_p,get_derivatives=False))[:,0:1]
    
    def coef_cholesky(self,model,X,Y,**kwargs):
        " Calculate the coefficients by using Cholesky decomposition. "
        # Calculate the kernel with and without noise
        KXX_n,KXX,n_data=self.kxx_reg(model,X)
        # Cholesky decomposition
        L,low=cho_factor(KXX_n)
        # Subtract the prior mean to the training target
        Y_p=self.y_prior(X,Y,model,L=L,low=low)
        # Get the coefficients
        coef=cho_solve((L,low),Y_p,check_finite=False)
        return coef,L,low,Y_p,KXX,n_data

    def get_eig(self,model,X,Y,**kwargs):
        " Calculate the eigenvalues. " 
        # Calculate the kernel with and without noise
        KXX,n_data=self.kxx_corr(model,X)
        # Eigendecomposition
        try:
            D,U=eigh(KXX)
        except Exception as e:
            import logging
            import scipy.linalg
            logging.error("An error occurred: %s", str(e))
            # More robust but slower eigendecomposition
            D,U=scipy.linalg.eigh(KXX,driver='ev')
        # Subtract the prior mean to the training target
        Y_p=self.y_prior(X,Y,model,D=D,U=U)
        UTY=(np.matmul(U.T,Y_p)).reshape(-1)**2
        return D,U,Y_p,UTY,KXX,n_data
    
    def get_cinv(self,model,X,Y,**kwargs):
        " Get the inverse covariance matrix. "
        coef,L,low,Y_p,KXX,n_data=self.coef_cholesky(model,X,Y)
        cinv=cho_solve((L,low),np.identity(n_data),check_finite=False)
        return coef,cinv,Y_p,KXX,n_data
    
    def logpriors(self,hp,pdis=None,jac=False,**kwargs):
        " Log of the prior distribution value for the hyperparameters. "
        # If no prior distribution is used for the hyperparameters
        if pdis is None:
            return 0.0
        # If the log probability is calculated
        if not jac:
            lprior=0.0
            for para,value in hp.items():
                if para in pdis.keys():
                    lprior=lprior+pdis[para].ln_pdf(value)
            if isinstance(lprior,float):
                return lprior
            return lprior.reshape(-1)
        # If the derivate of the log probability is calculated wrt. the hyperparameters
        lprior_deriv=np.array([])
        for para,value in hp.items():
            if para in pdis.keys():
                lprior_deriv=np.append(lprior_deriv,np.array(pdis[para].ln_deriv(value)).reshape(-1))
            else:
                lprior_deriv=np.append(lprior_deriv,np.zeros((len(value))))
        return lprior_deriv

    def get_K_inv_deriv(self,K_deriv,KXX_inv,**kwargs):
        " Get the diagonal elements of the matrix product of the inverse and derivative covariance matrix. "
        return np.einsum('ij,dji->d',KXX_inv,K_deriv)
    
    def get_K_deriv(self,model,parameter,X,KXX,**kwargs):
        " Get the gradient of the covariance matrix wrt. the hyperparameter. "
        K_deriv=model.get_gradients(X,[parameter],KXX=KXX)[parameter]
        return K_deriv
    
    def get_prefactor2(self,model,**kwargs):
        " Get the prefactor hyperparameter in log space and the squared in linear space (exp). "
        prefactor=model.hp['prefactor'][0]
        return prefactor,np.exp(2.0*prefactor)
    
    def get_prior_parameters(self,model,**kwargs):
        " Get the prior parameters. "
        return model.prior.get_parameters()
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(get_prior_mean=self.get_prior_mean)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs
    
    def copy(self):
        " Copy the object. "
        # Get all arguments
        arg_kwargs,constant_kwargs,object_kwargs=self.get_arguments()
        # Make a clone
        clone=self.__class__(**arg_kwargs)
        # Check if constants have to be saved
        if len(constant_kwargs.keys()):
            for key,value in constant_kwargs.items():
                clone.__dict__[key]=value
        # Check if objects have to be saved
        if len(object_kwargs.keys()):
            for key,value in object_kwargs.items():
                clone.__dict__[key]=value.copy()
        return clone
    
    def __repr__(self):
        arg_kwargs=self.get_arguments()[0]
        str_kwargs=",".join([f"{key}={value}" for key,value in arg_kwargs.items()])
        return "{}({})".format(self.__class__.__name__,str_kwargs)
    


import numpy as np
from .constant import Prior_constant

class Prior_min(Prior_constant):
    def __init__(self,yp=0.0,add=0.0,**kwargs):
        """
        The prior mean of the targets. 
        The prior mean is used as a baseline of the target values.
        The prior mean is the minimum of the target value if given else it is 0. 
        A value can be added to the constant.
        Parameters:
            yp : float
                The prior mean constant
            add : float
                A value added to the found prior mean from data.
        """
        self.update_arguments(yp=yp,add=add,**kwargs)
    
    def update(self,X,Y,**kwargs):
        self.update_arguments(yp=np.min(Y[:,0]))
        return self



import numpy as np
from scipy.linalg import cho_solve
from .objectivefunction_gpatom import ObjectiveFuctionGPAtom

class LogLikelihood(ObjectiveFuctionGPAtom):
    def __init__(self,get_prior_mean=False,**kwargs):
        """ 
        The log-likelihood objective function that is used to optimize the hyperparameters. 
        Parameters:
            get_prior_mean: bool
                Whether to save the parameters of the prior mean in the solution.
        """
        super().__init__(get_prior_mean=get_prior_mean,**kwargs)
    
    def function(self,theta,parameters,model,X,Y,pdis=None,jac=False,**kwargs):
        hp,parameters_set=self.make_hp(theta,parameters)
        model=self.update_model(model,hp)
        coef,L,low,Y_p,KXX,n_data=self.coef_cholesky(model,X,Y)
        prefactor,prefactor2=self.get_prefactor2(model)
        nlp=0.5*np.matmul(Y_p.T,coef).item(0)/prefactor2+n_data*prefactor+np.sum(np.log(np.diagonal(L)))+0.5*n_data*np.log(2.0*np.pi)
        nlp=nlp-self.logpriors(hp,pdis,jac=False)
        if jac:
            return nlp,self.derivative(hp,parameters_set,model,X,Y_p,KXX,L,low,coef,prefactor2,n_data,pdis,**kwargs)   
        return nlp
    
    def derivative(self,hp,parameters_set,model,X,Y_p,KXX,L,low,coef,prefactor2,n_data,pdis,**kwargs):
        nlp_deriv=np.array([])
        KXX_inv=cho_solve((L,low),np.identity(n_data),check_finite=False)
        for para in parameters_set:
            if para=='prefactor':
                nlp_deriv=np.append(nlp_deriv,-np.matmul(Y_p.T,coef).item(0)/prefactor2+n_data)
                continue
            K_deriv=self.get_K_deriv(model,para,X=X,KXX=KXX)
            K_deriv_cho=self.get_K_inv_deriv(K_deriv,KXX_inv)
            nlp_deriv=np.append(nlp_deriv,-0.5*np.matmul(coef.T,np.matmul(K_deriv,coef)).reshape(-1)/prefactor2+0.5*K_deriv_cho)
        nlp_deriv=nlp_deriv-self.logpriors(hp,pdis,jac=True)
        return nlp_deriv



from .optimizer import Optimizer
import numpy as np

class LocalOptimizer(Optimizer):
    def __init__(self,maxiter=5000,jac=True,tol=1e-3,**kwargs):
        """
        The local optimizer used for optimzing the objective function wrt. the hyperparameters.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            jac : bool
                Whether to use the gradient of the objective function wrt. the hyperparameters.
            tol : float
                A tolerance criterion for convergence. 
        """
        # This optimizer can not be parallelized
        self.parallel=False
        # Set all the arguments
        self.update_arguments(maxiter=maxiter,
                              jac=jac,
                              tol=tol,
                              **kwargs)

    def run(self,func,theta,parameters,model,X,Y,pdis,**kwargs):
        raise NotImplementedError()

    def update_arguments(self,maxiter=None,jac=None,tol=None,**kwargs):
        """
        Update the optimizer with its arguments. The existing arguments are used if they are not given.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            jac : bool
                Whether to use the gradient of the objective function wrt. the hyperparameters.
            tol : float
                A tolerance criterion for convergence. 
        Returns:
            self: The updated object itself.
        """
        if maxiter is not None:
            self.maxiter=int(maxiter)
        if jac is not None:
            self.jac=jac
        if tol is not None:
            self.tol=tol
        return self
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(maxiter=self.maxiter,jac=self.jac,tol=self.tol)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs
    

class ScipyOptimizer(LocalOptimizer):
    def __init__(self,maxiter=5000,jac=True,tol=1e-8,method='l-bfgs-b',bounds=None,use_bounds=False,options={},opt_kwargs={},**kwargs):
        """
        The local optimizer used for optimzing the objective function wrt. the hyperparameters.
        This method uses the SciPy minimizers. 
        (https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html)
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            jac : bool
                Whether to use the gradient of the objective function wrt. the hyperparameters.
            tol : float
                A tolerance criterion for convergence. 
            method : str
                The minimizer method used in SciPy.
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
                All global optimization methods are using boundary conditions. 
            use_bounds : bool
                Whether to use the boundary conditions or not.
                Only some methods can use boundary conditions.
            options : dict
                Solver options used in the SciPy minimizer.
            opt_kwargs : dict
                Extra arguments used in the SciPy minimizer.
        """
        # This optimizer can not be parallelized
        self.parallel=False
        # Set boundary conditions
        self.bounds=None
        # Set options
        self.options={}
        # Set optimization arguments
        self.opt_kwargs={}
        # Set all the arguments
        self.update_arguments(maxiter=maxiter,
                              jac=jac,
                              tol=tol,
                              method=method,
                              bounds=bounds,
                              use_bounds=use_bounds,
                              options=options,
                              opt_kwargs=opt_kwargs,
                              **kwargs)

    def run(self,func,theta,parameters,model,X,Y,pdis,**kwargs):
        from scipy.optimize import minimize
        # Get the objective function arguments
        func_args=self.get_func_arguments(parameters,model,X,Y,pdis,self.jac)
        # Get bounds or set it to default argument
        if self.use_bounds:
            bounds=self.make_bounds(parameters,array=True)
        else:
            bounds=None
        # Minimize objective function with SciPy
        sol=minimize(self.get_fun(func),x0=theta,
                     method=self.method,jac=self.jac,
                     tol=self.tol,args=func_args,
                     bounds=bounds,
                     options=self.options,**self.opt_kwargs)
        return self.get_final_solution(sol,func,parameters,model,X,Y,pdis)

    def update_arguments(self,maxiter=None,jac=None,tol=None,method=None,bounds=None,use_bounds=None,options=None,opt_kwargs=None,**kwargs):
        """
        Update the optimizer with its arguments. The existing arguments are used if they are not given.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            jac : bool
                Whether to use the gradient of the objective function wrt. the hyperparameters.
            tol : float
                A tolerance criterion for convergence. 
            method : str
                The minimizer method used in SciPy.
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
                All global optimization methods are using boundary conditions. 
            use_bounds : bool
                Whether to use the boundary conditions or not.
                Only some methods can use boundary conditions.
            options : dict
                Solver options used in the SciPy minimizer.
            opt_kwargs : dict
                Extra arguments used in the SciPy minimizer.
        Returns:
            self: The updated object itself.
        """
        if jac is not None:
            self.jac=jac
        if tol is not None:
            self.tol=tol
        if method is not None:
            self.method=method.lower()
            # If method is updated then maxiter must be updated
            if maxiter is None:
                maxiter=self.maxiter
        if options is not None:
            self.options.update(options)
        if maxiter is not None:
            self.maxiter=int(maxiter)
            if self.method in ['nelder-mead']:
                self.options['maxfev']=self.maxiter
            elif self.method in ['l-bfgs-b','tnc']:
                self.options['maxfun']=self.maxiter
            else:
                self.options['maxiter']=self.maxiter
        if bounds is not None:
            self.bounds=bounds.copy()
        if use_bounds is not None:
            if self.bounds is not None and self.method in ['nelder-mead','l-bfgs-b','tnc','slsqp','powell','trust-constr','cobyla']:
                self.use_bounds=use_bounds
            else:
                self.use_bounds=False
        if opt_kwargs is not None:
            self.opt_kwargs.update(opt_kwargs)
        return self
    
    def make_bounds(self,parameters,array=True,**kwargs):
        " Make the boundary conditions of the hyperparameters. "
        return self.bounds.get_bounds(parameters=parameters,array=array,**kwargs)
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(maxiter=self.maxiter,jac=self.jac,tol=self.tol,
                        method=self.method,bounds=self.bounds,
                        use_bounds=self.use_bounds,options=self.options,
                        opt_kwargs=self.opt_kwargs)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs


class ScipyPriorOptimizer(ScipyOptimizer):
    def __init__(self,maxiter=5000,jac=True,tol=1e-8,method='l-bfgs-b',bounds=None,use_bounds=False,options={},opt_kwargs={},**kwargs):
        """
        The local optimizer used for optimzing the objective function wrt. the hyperparameters.
        This method uses the SciPy minimizers. 
        (https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html)
        If prior distributions of the hyperparameters are used, it will start by include
        the prior distributions and then restart with excluded prior distributions.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            jac : bool
                Whether to use the gradient of the objective function wrt. the hyperparameters.
            tol : float
                A tolerance criterion for convergence. 
            method : str
                The minimizer method used in SciPy.
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
                All global optimization methods are using boundary conditions. 
            use_bounds : bool
                Whether to use the boundary conditions or not.
                Only some methods can use boundary conditions.
            options : dict
                Solver options used in the SciPy minimizer.
            opt_kwargs : dict
                Extra arguments used in the SciPy minimizer.
        """
        super().__init__(maxiter=maxiter,jac=jac,tol=tol,method=method,
                         bounds=bounds,use_bounds=use_bounds,options=options,
                         opt_kwargs=opt_kwargs,**kwargs)
        
    def run(self,func,theta,parameters,model,X,Y,pdis,**kwargs):
        # Check if prior distributions of the hyperparameters are used
        if pdis is None:
            return super().run(func,theta,parameters,model,X,Y,pdis,**kwargs)
        # If the prior distributions of the hyperparameters are used then include them in the optimization
        sol=super().run(func,theta,parameters,model,X,Y,pdis,**kwargs)
        # Save the number of evaluations and the new best hyperparameters
        nfev=sol['nfev']
        # Exclude the prior distributions of the hyperparameters in the optimization
        sol=super().run(func,sol['x'],parameters,model,X,Y,None,**kwargs)
        sol['nfev']+=nfev
        sol['nit']=2
        return sol
    

class ScipyGuessOptimizer(ScipyOptimizer):
    def __init__(self,maxiter=5000,jac=True,tol=1e-8,method='l-bfgs-b',bounds=None,use_bounds=False,options={},opt_kwargs={},**kwargs):
        """
        The local optimizer used for optimzing the objective function wrt. the hyperparameters.
        This method uses the SciPy minimizers. 
        (https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html)
        Use boundary conditions to give an extra guess of the hyperparameters that also are optimized.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            jac : bool
                Whether to use the gradient of the objective function wrt. the hyperparameters.
            tol : float
                A tolerance criterion for convergence. 
            method : str
                The minimizer method used in SciPy.
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
                All global optimization methods are using boundary conditions. 
            use_bounds : bool
                Whether to use the boundary conditions or not.
                Only some methods can use boundary conditions.
            options : dict
                Solver options used in the SciPy minimizer.
            opt_kwargs : dict
                Extra arguments used in the SciPy minimizer.
        """
        super().__init__(maxiter=maxiter,jac=jac,tol=tol,method=method,
                         bounds=bounds,use_bounds=use_bounds,options=options,
                         opt_kwargs=opt_kwargs,**kwargs)
        
    def run(self,func,theta,parameters,model,X,Y,pdis,**kwargs):
        # Check if boundary conditions of the hyperparameters are used
        if self.bounds is None:
            return super().run(func,theta,parameters,model,X,Y,pdis,**kwargs)
        # Use the boundary conditions to give an educated guess of the hyperparmeters
        theta_guess=self.guess_hp(parameters,array=True)
        sol_ed=super().run(func,theta_guess,parameters,model,X,Y,pdis,**kwargs)
        # Optimize the initial hyperparameters
        sol=super().run(func,theta,parameters,model,X,Y,pdis,**kwargs)
        # Update the solution if it is better
        sol=self.compare_solutions(sol,sol_ed)
        sol['nit']=2
        return sol
    
    def guess_hp(self,parameters,array=True,**kwargs):
        " Make a guess of the hyperparameters from the boundary conditions. "
        return self.bounds.get_hp(parameters=parameters,array=array,**kwargs)




from .prior import Prior
from .constant import Prior_constant
from .mean import Prior_mean
from .median import Prior_median
from .min import Prior_min
from .max import Prior_max
from .first import Prior_first

__all__ = ["Prior","Prior_constant","Prior_mean","Prior_median","Prior_min","Prior_max","Prior_first"]


class Prior:
    def __init__(self,**kwargs):
        """
        The prior mean of the targets. 
        The prior mean is used as a baseline of the target values.
        """
        self.update_arguments(**kwargs)
    
    def get(self,X,Y,get_derivatives=True,**kwargs):
        """
        Get the prior mean of the targets.
        Parameters:
            features : (N,D) array or (N) list of fingerprint objects
                Training features with N data points.
            targets : (N,1) array or (N,1+D) array
                Training targets with N data points.
                If get_derivatives=True, the training targets is in first column and derivatives is in the next columns.
            get_derivatives: bool
                Whether to give the prior mean of the derivatives of targets.
        """
        raise NotImplementedError()
    
    def update(self,X,Y,**kwargs):
        """
        Update the prior mean with the given data.
        Parameters:
            features : (N,D) array or (N) list of fingerprint objects
                Training features with N data points.
            targets : (N,1) array or (N,1+D) array
                Training targets with N data points.
                If get_derivatives=True, the training targets is in first column and derivatives is in the next columns.
        Returns:
            self: The updated object itself.
        """
        self.update_arguments()
        return self
    
    def get_parameters(self,**kwargs):
        """
        Get the prior mean parameters.
        Returns:
            dict: A dictionary with the parameters used in the prior mean.
        """
        return dict()
    
    def update_arguments(self,**kwargs):
        """
        Update the class with its arguments. The existing arguments are used if they are not given.
        Returns:
            self: The updated object itself.
        """
        return self
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict()
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs
    
    def copy(self):
        " Copy the object. "
        # Get all arguments
        arg_kwargs,constant_kwargs,object_kwargs=self.get_arguments()
        # Make a clone
        clone=self.__class__(**arg_kwargs)
        # Check if constants have to be saved
        if len(constant_kwargs.keys()):
            for key,value in constant_kwargs.items():
                clone.__dict__[key]=value
        # Check if objects have to be saved
        if len(object_kwargs.keys()):
            for key,value in object_kwargs.items():
                clone.__dict__[key]=value.copy()
        return clone
    
    def __repr__(self):
        arg_kwargs=self.get_arguments()[0]
        str_kwargs=",".join([f"{key}={value}" for key,value in arg_kwargs.items()])
        return "{}({})".format(self.__class__.__name__,str_kwargs)



import numpy as np

class HPBoundaries:
    def __init__(self,bounds_dict={},scale=1.0,log=True,**kwargs):
        """ 
        Boundary conditions for the hyperparameters.
        A dictionary with boundary conditions of the hyperparameters can be given as an argument.
        Machine precisions are used as boundary conditions for the hyperparameters not given in the dictionary.
        Parameters:
            bounds_dict : dict
                A dictionary with boundary conditions as numpy (H,2) arrays with two columns for each type of hyperparameter.
            scale : float
                Scale the boundary conditions.
            log : bool
                Whether to use hyperparameters in log-scale or not.
        """
        self.update_arguments(bounds_dict=bounds_dict,
                              scale=scale,
                              log=log,
                              **kwargs)

    def update_bounds(self,model,X,Y,parameters,**kwargs):
        """ 
        Create and update the boundary conditions for the hyperparameters. 
        Parameters:
            model : Model
                The Machine Learning Model with kernel and prior that are optimized.
            X : (N,D) array
                Training features with N data points and D dimensions.
            Y : (N,1) array or (N,D+1) array
                Training targets with or without derivatives with N data points.
            parameters : (H) list of strings
                A list of names of the hyperparameters.
        Returns:
            self : The object itself.
        """
        # Update parameters
        self.make_parameters_set(parameters)
        # Make bounds (Parameters and keys in bounds_dict have to be the same)
        self.bounds_dict=self.make_bounds(model,X,Y,parameters,self.parameters_set)
        return self

    def get_bounds(self,parameters=None,array=False,**kwargs):
        """ 
        Get the boundary conditions of the hyperparameters. 
        Parameters :
            parameters : list of str or None
                A list of the specific used hyperparameter names as strings.
                If parameters=None, then the stored hyperparameters are used.
            array : bool
                Whether to get an array or a dictionary as output.
        Returns:
            (H,2) array : The boundary conditions as an array if array=True.
            or
            dict : A dictionary of the boundary conditions. 
        """
        # Make the sorted unique hyperparameters if they are given
        parameters_set=self.get_parameters_set(parameters=parameters)
        # Make the boundary conditions for the given hyperparameters
        if array:
            return np.concatenate([self.bounds_dict[para] for para in parameters_set],axis=0)
        return {para:self.bounds_dict[para].copy() for para in parameters_set}
    
    def get_hp(self,parameters=None,array=False,**kwargs):
        """ 
        Get the guess of the hyperparameters. 
        The mean of the boundary conditions in log-space is used as the guess. 
        Parameters:
            parameters : list of str or None
                A list of the specific used hyperparameter names as strings.
                If parameters=None, then the stored hyperparameters are used.
            array : bool
                Whether to get an array or a dictionary as output.
        Returns:
            (H) array : The guesses of the hyperparameters as an array if array=True.
            or
            dict : A dictionary of the guesses of the hyperparameters. 
        """
        # Make the sorted unique hyperparameters if they are given
        parameters_set=self.get_parameters_set(parameters=parameters)
        if self.log:
            if array:
                return np.concatenate([np.mean(self.bounds_dict[para],axis=1) for para in parameters_set])
            return {para:np.mean(self.bounds_dict[para],axis=1) for para in parameters_set}
        if array:
            return np.concatenate([np.exp(np.mean(np.log(self.bounds_dict[para]),axis=1)) for para in parameters_set])
        return {para:np.exp(np.mean(np.log(self.bounds_dict[para]),axis=1)) for para in parameters_set}
    
    def make_lines(self,parameters=None,ngrid=80,**kwargs):
        """ 
        Make grid in each dimension of the hyperparameters from the boundary conditions.
        Parameters:
            parameters : list of str or None
                A list of the specific used hyperparameter names as strings.
                If parameters=None, then the stored hyperparameters are used.
            ngrid : int or (H) list
                An integer or a list with number of grid points in each dimension.
        Returns:
            (H,) list : A list with grid points for each (H) hyperparameters.
        """
        bounds=self.get_bounds(parameters=parameters,array=True)
        if isinstance(ngrid,(int,float)):
            ngrid=[int(ngrid)]*len(bounds)
        return [np.linspace(bound[0],bound[1],ngrid[b]) for b,bound in enumerate(bounds)]
    
    def make_single_line(self,parameter,ngrid=80,i=0,**kwargs):
        """ 
        Make grid in one dimension of the hyperparameters from the boundary conditions.
        Parameters:
            parameters : str
                A string of the hyperparameter name.
            ngrid : int
                An integer with number of grid points in each dimension.
            i : int
                The index of the hyperparameter used if multiple hyperparameters of the same type exist. 
        Returns: 
            (ngrid) array : A grid of ngrid points for the given hyperparameter.
        """
        if not isinstance(ngrid,(int,float)):
            ngrid=ngrid[int(self.parameters.index(parameter)+i)]
        bound=self.bounds_dict[parameter][i]
        return np.linspace(bound[0],bound[1],int(ngrid))
    
    def sample_thetas(self,parameters=None,npoints=50,**kwargs):
        """ 
        Sample hyperparameters from the boundary conditions. 
        Parameters:
            parameters : list of str or None
                A list of the specific used hyperparameter names as strings.
                If parameters=None, then the stored hyperparameters are used.
            npoints : int
                Number of points to sample.   
        Returns:
            (npoints,H) array : An array with sampled hyperparameters. 
        """
        bounds=self.get_bounds(parameters=parameters,array=True)
        return np.random.uniform(low=bounds[:,0],high=bounds[:,1],size=(int(npoints),len(bounds)))
    
    def update_arguments(self,bounds_dict=None,scale=None,log=None,**kwargs):
        """
        Update the class with its arguments. The existing arguments are used if they are not given.
        Parameters:
            bounds_dict : dict
                A dictionary with boundary conditions as numpy (H,2) arrays with two columns for each type of hyperparameter.
            scale : float
                Scale the boundary conditions.
            log : bool
                Whether to use hyperparameters in log-scale or not.
        Returns:
            self: The updated object itself.
        """
        if bounds_dict is not None:
            self.initiate_bounds_dict(bounds_dict)
        if scale is not None:
            self.scale=scale
        if log is not None:
            self.log=log
        return self
    
    def make_bounds(self,model,X,Y,parameters,parameters_set,**kwargs):
        " Make the boundary conditions with educated guesses of the length-scale hyperparamer(s). "
        eps_lower,eps_upper=self.get_boundary_limits()
        bounds={}
        for para in parameters_set:
            if para in self.bounds_dict:
                bounds[para]=self.bounds_dict[para].copy()
            else:
                bounds[para]=np.full((parameters.count(para),2),[eps_lower,eps_upper])
        return bounds
    
    def initiate_bounds_dict(self,bounds_dict,**kwargs):
        " Make and store the hyperparameter types and the dictionary with hyperparameter bounds. "
        # Copy the boundary condition values
        self.bounds_dict={key:np.array(value) for key,value in bounds_dict.items()}
        if 'correction' in self.bounds_dict.keys():
            self.bounds_dict.pop('correction')
        # Extract the hyperparameter names
        self.parameters_set=sorted(bounds_dict.keys())
        self.parameters=sum([[para]*len(bounds_dict[para]) for para in self.parameters_set],[])
        return self
    
    def make_parameters_set(self,parameters,**kwargs):
        " Make and store the hyperparameters types. "
        parameters=list(parameters)
        if 'correction' in parameters:
            parameters.remove('correction')
        self.parameters=sorted(parameters)
        self.parameters_set=sorted(set(parameters))
        return self.parameters_set

    def get_parameters_set(self,parameters=None,**kwargs):
        " Make the sorted unique hyperparameters if they are given else get the stored. "
        # Get the sorted and unique stored hyperparameter sets
        if parameters is None:
            return self.parameters_set.copy()
        # Make the sorted unique hyperparameters if they are given
        return sorted(set(parameters))
    
    def get_n_parameters(self,parameters=None,**kwargs):
        " Get the number of hyperparameters used or stored. "
        if parameters is None:
            return len(self.parameters)
        return len(parameters)
    
    def get_boundary_limits(self,**kwargs):
        " Get the machine precision limits for the hyperparameters. "
        eps_lower=10*np.sqrt(2.0*np.finfo(float).eps)/self.scale
        if self.log:
            eps_lower=np.log(eps_lower)
            return eps_lower,-eps_lower
        return eps_lower,1.0/eps_lower
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(bounds_dict=self.bounds_dict,
                        scale=self.scale,
                        log=self.log)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs

    def copy(self):
        " Copy the object. "
        # Get all arguments
        arg_kwargs,constant_kwargs,object_kwargs=self.get_arguments()
        # Make a clone
        clone=self.__class__(**arg_kwargs)
        # Check if constants have to be saved
        if len(constant_kwargs.keys()):
            for key,value in constant_kwargs.items():
                clone.__dict__[key]=value
        # Check if objects have to be saved
        if len(object_kwargs.keys()):
            for key,value in object_kwargs.items():
                clone.__dict__[key]=value.copy()
        return clone
    
    def __repr__(self):
        arg_kwargs=self.get_arguments()[0]
        str_kwargs=",".join([f"{key}={value}" for key,value in arg_kwargs.items()])
        return "{}({})".format(self.__class__.__name__,str_kwargs)



def update_pdis(model,parameters,X,Y,bounds=None,pdis=None,**kwargs):
    " Update given prior distribution of hyperparameters from educated guesses in log space. "
    # Check if prior distributions are given
    if pdis is None:
        return pdis
    # Make boundary conditions for updating the prior distributions
    if bounds is None:
        # Use strict educated guesses for the boundary conditions if not given
        from ..hpboundary.strict import StrictBoundaries
        bounds=StrictBoundaries(log=True,use_prior_mean=True)
        # Update boundary conditions to the data
        bounds.update_bounds(model,X,Y,parameters)
    # Make prior distributions for hyperparameters from boundary conditions
    for para,bound in bounds.get_bounds().items():
        if para in pdis.keys():
            # Use given prior distribution to update it 
            pdis[para]=pdis[para].min_max(bound[:,0],bound[:,1])
    return pdis



import numpy as np
from .boundary import HPBoundaries

class VariableTransformation(HPBoundaries):
    def __init__(self,var_dict={},bounds=None,s=0.14,eps=np.finfo(float).eps,**kwargs):
        """ 
        Make variable transformation of hyperparameters into an interval of (0,1). 
        A dictionary of mean and standard deviation values are used to make Logistic transformations.
        Boundary conditions can be used to calculate the variable transformation parameters.
        Parameters:
            var_dict : dict
                A dictionary with the variable transformation parameters (mean,std) for each hyperparameter.
            bounds : Boundary condition class 
                A Boundary condition class that make the boundaries of the hyperparameters. 
                The boundaries are used to calculate the variable transformation parameters. 
            s : float
                The scale parameter in a Logistic distribution. 
                It determines how large part of the distribution that is within the boundaries. 
                s=0.5*p/(ln(p)-ln(1-p)) with p being the quantile that the boundaries constitute.
            eps : float
                The first value of a grid in the variable transformed hyperparameter space.
                The last value of a grid is 1.0-eps. 
        """
        # Set the default boundary conditions
        if bounds is None:
            from .strict import StrictBoundaries
            bounds=StrictBoundaries(bounds_dict={},scale=1.0,log=True,use_prior_mean=True)
        # Set all the arguments
        self.update_arguments(var_dict=var_dict,
                              bounds=bounds,
                              s=s,
                              eps=eps,
                              **kwargs)


    def update_bounds(self,model,X,Y,parameters,**kwargs):
        """ 
        Create and update the boundary conditions for the hyperparameters. 
        Therefore the variable transformation parameters are also updated. 
        Parameters:
            model : Model
                The Machine Learning Model with kernel and prior that are optimized.
            X : (N,D) array
                Training features with N data points and D dimensions.
            Y : (N,1) array or (N,D+1) array
                Training targets with or without derivatives with N data points.
            parameters : (H) list of strings
                A list of names of the hyperparameters.
        Returns:
            self : The object itself.
        """
        # Update the parameters used
        self.make_parameters_set(parameters)
        # Update the boundary conditions and get them
        self.bounds.update_bounds(model,X,Y,parameters)
        self.bounds_dict=self.bounds.get_bounds(array=False)
        # Update the variable transformation parameters
        for para,bound in self.bounds_dict.items():
            self.var_dict[para]={'mean':np.mean(bound,axis=1),
                                 'std':self.s*np.abs(bound[:,1]-bound[:,0])}
        return self
    
    def get_variable_transformation_parameters(self,parameters=None,array=False,**kwargs):
        """ 
        Get the variable transformation parameters. 
        Parameters:
            parameters : list of str or None
                A list of the specific used hyperparameter names as strings.
                If parameters=None, then the stored hyperparameters are used.
            array : bool
                Whether to get an array for the mean and std or a dictionary as output.
        Returns:
            dict : A dictionary of the variable transformation parameters. 
            If array=True, a dictionary with mean and std is given instead.
        """
        # Make the sorted unique hyperparameters if they are given
        parameters_set=self.get_parameters_set(parameters=parameters)
        if array:
            var_dict_array={}
            var_dict_array['mean']=np.concatenate([self.var_dict[para]['mean'] for para in parameters_set],axis=0)
            var_dict_array['std']=np.concatenate([self.var_dict[para]['std'] for para in parameters_set],axis=0)
            return var_dict_array
        return {para:self.var_dict[para].copy() for para in parameters_set}
    
    def transformation(self,hp,array=False,**kwargs):
        """ 
        Transform the hyperparameters with the variable transformation to get a dictionary. 
        Parameters:
            hp : dict
                The dictionary of the hyperparameters
            array : bool
                Whether to get an array or a dictionary as output.
        Returns:
            (H) array : The variable transformed hyperparameters as an array if array=True.
            or
            dict : A dictionary of the variable transformed hyperparameters. 
        """
        if array:
            return np.concatenate([self.transform(theta,self.var_dict[para]['mean'],self.var_dict[para]['std']) for para,theta in hp.items()])
        return {para:self.transform(theta,self.var_dict[para]['mean'],self.var_dict[para]['std']) for para,theta in hp.items()}
    
    def reverse_trasformation(self,t,array=False,**kwargs):
        """ 
        Transform the variable transformed hyperparameters back to the hyperparameters dictionary.
        Parameters:
            t : dict
                The dictionary of the variable transformed hyperparameters
            array : bool
                Whether to get an array or a dictionary as output.
        Returns:
            (H) array : The retransformed hyperparameters as an array if array=True.
            or
            dict : A dictionary of the retransformed hyperparameters. 
        """
        if array:
            return np.concatenate([self.retransform(ti,self.var_dict[para]['mean'],self.var_dict[para]['std']) for para,ti in t.items()])
        return {para:self.retransform(ti,self.var_dict[para]['mean'],self.var_dict[para]['std']) for para,ti in t.items()}
    
    def get_bounds(self,parameters=None,array=False,transformed=False,**kwargs):
        """
        Get the boundary conditions of hyperparameters. 
        Parameters :
            parameters : list of str or None
                A list of the specific used hyperparameter names as strings.
                If parameters=None, then the stored hyperparameters are used.
            array : bool
                Whether to get an array or a dictionary as output.
            transformed : bool
                If transformed=True, the boundaries is in variable transformed space.
                If transformed=False, the boundaries is transformed back to hyperparameter space.
        Returns:
            (H,2) array : The boundary conditions as an array if array=True.
            or
            dict : A dictionary of the boundary conditions. 
        """
        # Get the bounds in the variable transformed space 
        if transformed:
            if array:
                n_parameters=self.get_n_parameters(parameters=parameters)
                return np.full((n_parameters,2),[self.eps,1.00-self.eps])
            # Make the sorted unique hyperparameters if they are given
            parameters_set=self.get_parameters_set(parameters=parameters)
            return {para:np.full((len(self.bounds_dict[para]),2),[self.eps,1.00-self.eps]) for para in parameters_set}
        # Get the bounds in the hyperparameter space 
        return self.bounds.get_bounds(parameters=parameters,array=array)
    
    def get_hp(self,parameters=None,array=False,transformed=False,**kwargs):
        """ 
        Get the guess of the hyperparameters. 
        The mean of the boundary conditions in log-space is used as the guess. 
        Parameters:
            parameters : list of str or None
                A list of the specific used hyperparameter names as strings.
                If parameters=None, then the stored hyperparameters are used.
            array : bool
                Whether to get an array or a dictionary as output.
            transformed : bool
                If transformed=True, the hyperparameters is in variable transformed space.
                If transformed=False, the hyperparameters is transformed back to hyperparameter space.
        Returns:
            (H) array : The guesses of the hyperparameters as an array if array=True.
            or
            dict : A dictionary of the guesses of the hyperparameters. 
        """
        # Get the hyperparameter guess in the variable transformed space (so in the middle 0.5)
        if transformed:
            if array:
                n_parameters=self.get_n_parameters(parameters=parameters)
                return np.full((n_parameters),0.50)
            # Make the sorted unique hyperparameters if they are given
            parameters_set=self.get_parameters_set(parameters=parameters)
            return {para:np.full((len(self.bounds_dict[para])),0.50) for para in parameters_set}
        # Get the hyperparameter guess in the hyperparameter space 
        return self.bounds.get_hp(parameters=parameters,array=array)

    def make_lines(self,parameters=None,ngrid=80,transformed=False,**kwargs):
        """ 
        Make grid in each dimension of the hyperparameters from the boundary conditions.
        Parameters:
            ngrid : int or (H) list
                An integer or a list with number of grid points in each dimension.
            transformed : bool
                If transformed=True, the grid is in variable transformed space.
                If transformed=False, the grid is transformed back to hyperparameter space.
        Returns:
            (H,) list : A list with grid points for each (H) hyperparameters.
        """
        # Get the number of hyperparameters
        n_parameters=self.get_n_parameters(parameters=parameters)
        # Make sure that a list of number grid points is used
        if isinstance(ngrid,(int,float)):
            ngrid=[int(ngrid)]*n_parameters
        # The grid is made within the variable transformed hyperparameters
        if transformed:
            return [np.linspace(self.eps,1.00-self.eps,ngrid[i]) for i in range(n_parameters)]
        # The grid is made within the variable transformed hyperparameters and then transformed back
        var_dict_array=self.get_variable_transformation_parameters(parameters=parameters,array=True)
        lines=[]
        for i,(vt_mean,vt_std) in enumerate(zip(var_dict_array['mean'],var_dict_array['std'])):
            t_line=np.linspace(self.eps,1.00-self.eps,ngrid[i])
            lines.append(self.retransform(t_line,vt_mean,vt_std))
        return lines
    
    def make_single_line(self,parameter,ngrid=80,i=0,transformed=False,**kwargs):
        """ 
        Make grid in each dimension of the hyperparameters from the boundary conditions.
        Parameters:
            parameters : str
                A string of the hyperparameter name.
            ngrid : int
                An integer with number of grid points in each dimension.
            i : int
                The index of the hyperparameter used if multiple hyperparameters of the same type exist.
            transformed : bool
                If transformed=True, the grid is in variable transformed space.
                If transformed=False, the grid is transformed back to hyperparameter space.
        Returns: 
            (ngrid) array : A grid of ngrid points for the given hyperparameter.
        """
        # Make sure that a int of number grid points is used
        if not isinstance(ngrid,(int,float)):
            ngrid=ngrid[int(self.parameters.index(parameter)+i)]
        # The grid is made within the variable transformed hyperparameters
        t_line=np.linspace(self.eps,1.00-self.eps,ngrid)
        if transformed:
            return t_line
        # The grid is transformed back to hyperparameter space
        return self.retransform(t_line,self.var_dict[parameter]['mean'][i],self.var_dict[parameter]['std'][i])
    
    def sample_thetas(self,parameters=None,npoints=50,transformed=False,**kwargs):
        """ 
        Sample hyperparameters from the variable transformed hyperparameter space.
        The sampled variable transformed hyperparameters are then transformed back to the hyperparameter space. 
        Parameters:
            parameters : list of str or None
                A list of the specific used hyperparameter names as strings.
                If parameters=None, then the stored hyperparameters are used.
            npoints : int
                Number of points to sample.
            transformed : bool
                If transformed=True, the samples are in variable transformed space.
                If transformed=False, the samples are transformed back to hyperparameter space.
        Returns:
            (npoints,H) array : An array with sampled hyperparameters. 
        """
        # Get the number of hyperparameters
        n_parameters=self.get_n_parameters(parameters=parameters)
        # Sample the hyperparameters from the variable transformed hyperparameter space
        samples=np.random.uniform(low=self.eps,high=1.00-self.eps,size=(npoints,n_parameters))
        # The samples are made within the variable transformed hyperparameters
        if transformed:
            return samples
        # The samples are transformed back to hyperparameter space
        var_dict_array=self.get_variable_transformation_parameters(parameters=parameters,array=True)
        for i,(vt_mean,vt_std) in enumerate(zip(var_dict_array['mean'],var_dict_array['std'])):
            samples[:,i]=self.retransform(samples[:,i],vt_mean,vt_std)
        return samples

    def update_arguments(self,var_dict=None,bounds=None,s=None,eps=None,**kwargs):
        """
        Update the class with its arguments. The existing arguments are used if they are not given.
        Parameters:
            var_dict : dict
                A dictionary with the variable transformation parameters (mean,std) for each hyperparameter.
            bounds : Boundary condition class 
                A Boundary condition class that make the boundaries of the hyperparameters. 
                The boundaries are used to calculate the variable transformation parameters. 
            s : float
                The scale parameter in a Logistic distribution. 
                It determines how large part of the distribution that is within the boundaries. 
                s=0.5*p/(ln(p)-ln(1-p)) with p being the quantile that the boundaries constitute.
            eps : float
                The first value of a grid in the variable transformed hyperparameter space.
                The last value of a grid is 1.0-eps. 
        Returns:
            self: The updated object itself.
        """
        if var_dict is not None:
            self.initiate_var_dict(var_dict)
        if bounds is not None:
            self.initiate_bounds_dict(bounds)
        if s is not None:
            self.s=s
        if eps is not None:
            self.eps=eps
        return self
    
    def transform(self,theta,vt_mean,vt_std,**kwargs):
        " Transform the hyperparameters with the variable transformation. "
        return 1.0/(1.0+np.exp(-(theta-vt_mean)/vt_std))

    def retransform(self,ti,vt_mean,vt_std,**kwargs):
        " Transform the variable transformed hyperparameters back to the hyperparameters. "
        return self.numeric_limits(vt_std*np.log(ti/(1.00-ti))+vt_mean)

    def numeric_limits(self,value,dh=0.1*np.log(np.finfo(float).max)):
        " Replace hyperparameters if they are outside of the numeric limits in log-space. "
        return np.where(-dh<value,np.where(value<dh,value,dh),-dh)
    
    def initiate_var_dict(self,var_dict,**kwargs):
        " Make and store the hyperparameter types and the dictionary with transformation parameters. "
        # Copy the variable transformation parameters
        self.var_dict={key:{'mean':np.array(value['mean']),'std':np.array(value['std'])} for key,value in var_dict.items()}
        if 'correction' in self.var_dict.keys():
            self.var_dict.pop('correction')
        # Extract the hyperparameters
        self.parameters_set=sorted(var_dict.keys())
        self.parameters=sum([[para]*len(var_dict[para]) for para in self.parameters_set],[])
        return self
    
    def initiate_bounds_dict(self,bounds,**kwargs):
        " Make and store the hyperparameter bounds. "
        # Copy the boundary condition object
        self.bounds=bounds.copy()
        self.bounds_dict=self.bounds.get_bounds(array=False)
        # Make sure log-scale of the hyperparameters are used
        if self.bounds.log==False:
            raise Exception('The Variable Transformation need to use boundary conditions in the log-scale!')
        return self
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(var_dict=self.var_dict,
                        bounds=self.bounds,
                        s=self.s,
                        eps=self.eps)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs



from scipy.optimize import OptimizeResult
import numpy as np

class Optimizer:
    def __init__(self,maxiter=5000,jac=True,**kwargs):
        """
        The optimizer used for optimzing the objective function wrt. the hyperparameters.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            jac : bool
                Whether to use the gradient of the objective function wrt. the hyperparameters.
        """
        # This optimizer can not be parallelized
        self.parallel=False
        # Set all the arguments
        self.update_arguments(maxiter=maxiter,jac=jac,**kwargs)

    def run(self,func,theta,parameters,model,X,Y,pdis,**kwargs):
        """
        Run the optimization method by optimizing the objective function wrt. the hyperparameters.
        Parameters:
            func : ObjectiveFunction class object
                The objective function class that is used to calculate the value.
            theta : (H) array
                An array with the hyperparameter values.
            parameters : (H) list of strings
                A list of names of the hyperparameters.
            model : Model class object
                The Machine Learning Model with kernel and prior that are optimized.
            X : (N,D) array
                Training features with N data points and D dimensions.
            Y : (N,1) array or (N,D+1) array
                Training targets with or without derivatives with N data points.
            pdis : dict
                A dict of prior distributions for each hyperparameter type.
        Returns:
            dict : A solution dictionary with objective function value, optimized hyperparameters,
                success statement, and number of used evaluations.
        """
        raise NotImplementedError()

    def update_arguments(self,maxiter=None,jac=None,**kwargs):
        """
        Update the optimizer with its arguments. The existing arguments are used if they are not given.
        Parameters:
            maxiter : int
                The maximum number of evaluations or iterations the optimizer can use.
            jac : bool
                Whether to use the gradient of the objective function wrt. the hyperparameters.
        Returns:
            self: The updated object itself.
        """
        if maxiter is not None:
            self.maxiter=int(maxiter)
        if jac is not None:
            self.jac=jac
        return self
    
    def get_final_solution(self,sol,func,parameters,model,X,Y,pdis,**kwargs):
        " Get the final solution from the objective function. "
        if self.parallel:
            sol=self.get_final_solution_parallel(sol,func,parameters,model,X,Y,pdis,**kwargs)
        else:
            sol=func.get_solution(sol,parameters,model,X,Y,pdis)
        return OptimizeResult(**sol)
    
    def get_final_solution_parallel(self,sol,func,parameters,model,X,Y,pdis,**kwargs):
        " Get all final solutions from each function at each rank. "
        from ase.parallel import world,broadcast
        size=world.size
        fun_sol=func.get_stored_solution()
        sol=func.get_solution(sol,parameters,model,X,Y,pdis)
        fun_sols=[broadcast(fun_sol['fun'],root=r) for r in range(size)]
        rank_min=np.argmin(fun_sols)
        return broadcast(sol,root=rank_min)
    
    def get_empty_solution(self,**kwargs):
        " Get an empty solution without any function evaluations. "
        sol={'fun':np.inf,'x':np.array([]),
             'success':False,'nfev':0,'nit':0,
             'message':"No function value calculated."}
        return sol
    
    def get_initial_solution(self,theta,func,func_args=(),**kwargs):
        " Get a solution with the evaluation of the initial hyperparameters. "
        sol={'fun':np.inf,'x':theta,
             'success':False,'nfev':1,'nit':1,
             'message':"Function value is calculated."}
        fun=self.get_fun(func)
        if self.jac:
            sol['fun'],sol['jac']=fun(theta,*func_args)
        else:
            sol['fun']=fun(theta,*func_args)
        return sol
    
    def get_func_arguments(self,parameters,model,X,Y,pdis,jac,**kwargs):
        " Get the arguments needed for the objective function. "
        return (parameters,model,X,Y,pdis,jac)
    
    def get_fun(self,func,**kwargs):
        " Get the function that evaluates the objective function. "
        return func.function
    
    def calculate_values(self,thetas,func,func_args=(),**kwargs):
        " Calculate a list of values with a function. "
        if self.parallel:
            return self.calculate_values_parallel(thetas,func,func_args=func_args,**kwargs)
        return np.array([func.function(theta,*func_args) for theta in thetas])
    
    def calculate_values_parallel(self,thetas,func,func_args=(),**kwargs):
        " Calculate a list of values with a function in parallel. "
        from ase.parallel import world,broadcast
        rank,size=world.rank,world.size
        f_list=np.array([func.function(theta,*func_args) for t,theta in enumerate(thetas) if rank==t%size])
        return np.array([broadcast(f_list,root=r) for r in range(size)]).T.reshape(-1)
    
    def compare_solutions(self,sol1,sol2,**kwargs):
        " Compare two solutions and use the solution with lowest function value. "
        # Store the number of used iterations
        nfev=sol1['nfev']+sol2['nfev']
        nit=sol1['nit']+sol2['nit']
        # Compare function values
        if sol2['fun']<sol1['fun']:
            # Solution 2 is best
            sol2['nfev']=nfev
            sol2['nit']=nit
            return sol2
        # Solution 1 is best
        sol1['nfev']=nfev
        sol1['nit']=nit
        return sol1

    def make_hp(self,theta,parameters,**kwargs):
        " Make hyperparameter dictionary from lists. "
        theta,parameters=np.array(theta),np.array(parameters)
        parameters_set=sorted(set(parameters))
        hp={para_s:theta[parameters==para_s] for para_s in parameters_set}
        return hp

    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(maxiter=self.maxiter,jac=self.jac)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs

    def copy(self):
        " Copy the object. "
        # Get all arguments
        arg_kwargs,constant_kwargs,object_kwargs=self.get_arguments()
        # Make a clone
        clone=self.__class__(**arg_kwargs)
        # Check if constants have to be saved
        if len(constant_kwargs.keys()):
            for key,value in constant_kwargs.items():
                clone.__dict__[key]=value
        # Check if objects have to be saved
        if len(object_kwargs.keys()):
            for key,value in object_kwargs.items():
                clone.__dict__[key]=value.copy()
        return clone
    
    def __repr__(self):
        arg_kwargs=self.get_arguments()[0]
        str_kwargs=",".join([f"{key}={value}" for key,value in arg_kwargs.items()])
        return "{}({})".format(self.__class__.__name__,str_kwargs)


class FunctionEvaluation(Optimizer):
    def __init__(self,jac=True,**kwargs):
        """
        A method used for evaluating the objective function for the given hyperparameters.
        Parameters:
            jac : bool
                Whether to use the gradient of the objective function wrt. the hyperparameters.
        """
        # This optimizer can not be parallelized
        self.parallel=False
        # Set all the arguments
        self.update_arguments(jac=jac,**kwargs)

    def run(self,func,theta,parameters,model,X,Y,pdis,**kwargs):
        """
        Run the evaluation of the objective function wrt. the hyperparameters.
        Parameters:
            func : ObjectiveFunction class object
                The objective function class that is used to calculate the value.
            theta : (H) array
                An array with the hyperparameter values.
            parameters : (H) list of strings
                A list of names of the hyperparameters.
            model : Model class object
                The Machine Learning Model with kernel and prior that are optimized.
            X : (N,D) array
                Training features with N data points and D dimensions.
            Y : (N,1) array or (N,D+1) array
                Training targets with or without derivatives with N data points.
            pdis : dict
                A dict of prior distributions for each hyperparameter type.
        Returns:
            dict : A solution dictionary with objective function value, hyperparameters,
                success statement, and number of used evaluations.
        """
        func_args=self.get_func_arguments(parameters,model,X,Y,pdis,jac=self.jac,**kwargs)
        sol=self.get_initial_solution(theta,func,func_args=func_args)
        return self.get_final_solution(sol,func,parameters,model,X,Y,pdis)

    def update_arguments(self,maxiter=None,jac=None,**kwargs):
        """
        Update the class with its arguments. The existing arguments are used if they are not given.
        Parameters:
            jac : bool
                Whether to use the gradient of the objective function wrt. the hyperparameters.
        Returns:
            self: The updated object itself.
        """
        if jac is not None:
            self.jac=jac
        return self
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(jac=self.jac)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs
    



from .boundary import HPBoundaries
from .length import LengthBoundaries
from .restricted import RestrictedBoundaries
from .educated import EducatedBoundaries
from .strict import StrictBoundaries
from .hptrans import VariableTransformation

__all__ = ["HPBoundaries","LengthBoundaries","RestrictedBoundaries","EducatedBoundaries","StrictBoundaries",\
           "VariableTransformation"]


from .optimizer import Optimizer,FunctionEvaluation
from .globaloptimizer import GlobalOptimizer,RandomSamplingOptimizer,GridOptimizer,IterativeLineOptimizer,BasinOptimizer,AnneallingOptimizer,AnneallingTransOptimizer,FactorizedOptimizer
from .localoptimizer import LocalOptimizer,ScipyOptimizer,ScipyPriorOptimizer,ScipyGuessOptimizer
from .linesearcher import LineSearchOptimizer,GoldenSearch,FineGridSearch,TransGridSearch
from .noisesearcher import NoiseGrid,NoiseGoldenSearch,NoiseFineGridSearch,NoiseTransGridSearch

__all__ = ["Optimizer","FunctionEvaluation",
           "GlobalOptimizer","RandomSamplingOptimizer","GridOptimizer","IterativeLineOptimizer","BasinOptimizer","AnneallingOptimizer","AnneallingTransOptimizer","FactorizedOptimizer",
           "LocalOptimizer","ScipyOptimizer","ScipyPriorOptimizer","ScipyGuessOptimizer",
           "LineSearchOptimizer","GoldenSearch","FineGridSearch","TransGridSearch",
           "NoiseGrid","NoiseGoldenSearch","NoiseFineGridSearch","NoiseTransGridSearch"]



import numpy as np
from scipy.spatial.distance import pdist,squareform
from .boundary import HPBoundaries

class LengthBoundaries(HPBoundaries):
    def __init__(self,bounds_dict={},scale=1.0,log=True,max_length=True,use_derivatives=False,**kwargs):
        """ 
        Boundary conditions for the hyperparameters with educated guess for the length-scale hyperparameter.
        Machine precisions are used as default boundary conditions for the rest of the hyperparameters not given in the dictionary.
        Parameters:
            bounds_dict : dict
                A dictionary with boundary conditions as numpy (H,2) arrays with two columns for each type of hyperparameter.
            scale : float
                Scale the boundary conditions.
            log : bool
                Whether to use hyperparameters in log-scale or not.
            max_length : bool
                Whether to use the maximum scaling for the length-scale or use a more reasonable scaling.
            use_derivatives : bool
                Whether the derivatives of the target are used in the model. 
                The boundary conditions of the length-scale hyperparameter(s) will change with the use_derivatives. 
                The use_derivatives will be updated when update_bounds is called.
        """
        self.update_arguments(bounds_dict=bounds_dict,
                              scale=scale,
                              log=log,
                              max_length=max_length,
                              use_derivatives=use_derivatives,
                              **kwargs)

    def update_arguments(self,bounds_dict=None,scale=None,log=None,max_length=None,use_derivatives=None,**kwargs):
        """
        Update the class with its arguments. The existing arguments are used if they are not given.
        Parameters:
            bounds_dict : dict
                A dictionary with boundary conditions as numpy (H,2) arrays with two columns for each type of hyperparameter.
            scale : float
                Scale the boundary conditions.
            log : bool
                Whether to use hyperparameters in log-scale or not.
            max_length : bool
                Whether to use the maximum scaling for the length-scale or use a more reasonable scaling.
            use_derivatives : bool
                Whether the derivatives of the target are used in the model. 
                The boundary conditions of the length-scale hyperparameter(s) will change with the use_derivatives. 
                The use_derivatives will be updated when update_bounds is called.
        Returns:
            self: The updated object itself.
        """
        if bounds_dict is not None:
            self.initiate_bounds_dict(bounds_dict)
        if scale is not None:
            self.scale=scale
        if log is not None:
            self.log=log
        if max_length is not None:
            self.max_length=max_length
        if use_derivatives is not None:
            self.use_derivatives=use_derivatives
        return self
    
    def make_bounds(self,model,X,Y,parameters,parameters_set,**kwargs):
        " Make the boundary conditions with educated guesses of the length-scale hyperparamer(s). "
        eps_lower,eps_upper=self.get_boundary_limits()
        self.get_use_derivatives(model)
        bounds={}
        for para in parameters_set:
            if para=='length':
                bounds[para]=self.length_bound(X,parameters.count(para))
            elif para in self.bounds_dict:
                bounds[para]=self.bounds_dict[para].copy()
            else:
                bounds[para]=np.full((parameters.count(para),2),[eps_lower,eps_upper])
        return bounds
    
    def length_bound(self,X,l_dim,**kwargs):
        " Get the minimum and maximum ranges of the length-scale in the educated guess regime within a scale. "
        # Get the minimum and maximum machine precision for exponential terms
        exp_lower=np.sqrt(-1/np.log(np.finfo(float).eps))/self.scale
        exp_max=np.sqrt(-1/np.log(1-np.finfo(float).eps))*self.scale
        # Use a smaller maximum boundary if only one length-scale is used or specified
        if not self.max_length or l_dim==1:
            exp_max=2.0*self.scale
        # Scale the convergence if derivatives of targets are used
        if self.use_derivatives:
            exp_lower=exp_lower*0.05
        lengths=np.zeros((l_dim,2))
        # If only one features is given then end 
        if len(X)==1:
            lengths[:,0]=exp_lower
            lengths[:,1]=exp_max
            if self.log:
                return np.log(lengths)
            return lengths
        # Ensure that the features are a matrix
        if not isinstance(X[0],(list,np.ndarray)):
            X=np.array([fp.vector for fp in X])
        for d in range(l_dim):
            # Calculate distances 
            if l_dim==1:
                dis=pdist(X)
            else:
                dis=pdist(X[:,d:d+1])
            # Calculate the maximum length-scale
            dis_max=exp_max*np.max(dis)
            if dis_max==0.0:
                dis_min,dis_max=exp_lower,exp_max
            else:
                # Calculate the minimum length-scale from the nearest neighbor distance
                dis_min=exp_lower*np.median(self.nearest_neighbors(dis))
                if dis_min==0.0:
                    dis_min=exp_lower
            # Transform into log-scale if specified
            lengths[d,0],lengths[d,1]=dis_min,dis_max
        if self.log:
            return np.log(lengths)
        return lengths
    
    def nearest_neighbors(self,dis,**kwargs):
        " Nearest neighbor distance. "
        dis_matrix=squareform(dis)
        np.fill_diagonal(dis_matrix,np.inf)
        return np.min(dis_matrix,axis=1)
    
    def get_use_derivatives(self,model,**kwargs):
        " Get whether the derivatives of targets are used in the model. "
        self.use_derivatives=model.use_forces
        return self.use_derivatives
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(bounds_dict=self.bounds_dict,
                        scale=self.scale,
                        log=self.log,
                        max_length=self.max_length,
                        use_derivatives=self.use_derivatives)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs



import numpy as np
from scipy.spatial.distance import pdist
from .educated import EducatedBoundaries

class StrictBoundaries(EducatedBoundaries):
    def __init__(self,bounds_dict={},scale=1.0,log=True,use_derivatives=False,use_prior_mean=True,**kwargs):
        """ 
        Boundary conditions for the hyperparameters with educated guess for 
        the length-scale, relative-noise, and prefactor hyperparameters.
        Stricter boundary conditions are used for the length-scale hyperparameter.
        Machine precisions are used as boundary conditions for other hyperparameters not given in the dictionary.        
        Parameters:
            bounds_dict : dict
                A dictionary with boundary conditions as numpy (H,2) arrays with two columns for each type of hyperparameter.
            scale : float
                Scale the boundary conditions.
            log : bool
                Whether to use hyperparameters in log-scale or not.
            use_derivatives : bool
                Whether the derivatives of the target are used in the model. 
                The boundary conditions of the length-scale hyperparameter(s) will change with the use_derivatives. 
                The use_derivatives will be updated when update_bounds is called.
            use_prior_mean : bool
                Whether to use the prior mean to calculate the boundary of the prefactor hyperparameter.
                If use_prior_mean=False the minimum and maximum target differences are used as the boundary conditions. 
        """
        super().__init__(bounds_dict=bounds_dict,scale=scale,log=log,use_derivatives=use_derivatives,use_prior_mean=use_prior_mean,**kwargs)
    
    def length_bound(self,X,l_dim,**kwargs):
        " Get the minimum and maximum ranges of the length-scale in the educated guess regime within a scale. "
        # Get the minimum and maximum scaling factors
        exp_lower=0.2/self.scale
        exp_max=4.0*self.scale
        # Scale the convergence if derivatives of targets are used
        if self.use_derivatives:
            exp_lower=exp_lower*0.05
        lengths=np.zeros((l_dim,2))
        # If only one features is given then end 
        if len(X)==1:
            lengths[:,0]=exp_lower
            lengths[:,1]=exp_max
            if self.log:
                return np.log(lengths)
            return lengths
        # Ensure that the features are a matrix
        if not isinstance(X[0],(list,np.ndarray)):
            X=np.array([fp.vector for fp in X])
        for d in range(l_dim):
            # Calculate distances 
            if l_dim==1:
                dis=pdist(X)
            else:
                dis=pdist(X[:,d:d+1])
            # Calculate the maximum length-scale
            dis_max=exp_max*np.median(dis)
            if dis_max==0.0:
                dis_min,dis_max=exp_lower,exp_max
            else:
                # Calculate the minimum length-scale from the nearest neighbor distance
                dis_min=exp_lower*np.median(self.nearest_neighbors(dis))
                if dis_min==0.0:
                    dis_min=exp_lower
            # Transform into log-scale if specified
            lengths[d,0],lengths[d,1]=dis_min,dis_max
        if self.log:
            return np.log(lengths)
        return lengths
    


import numpy as np
from scipy.spatial.distance import pdist
from .restricted import RestrictedBoundaries

class EducatedBoundaries(RestrictedBoundaries):
    def __init__(self,bounds_dict={},scale=1.0,log=True,max_length=True,use_derivatives=False,use_prior_mean=True,**kwargs):
        """ 
        Boundary conditions for the hyperparameters with educated guess for 
        the length-scale, relative-noise, and prefactor hyperparameters.
        Machine precisions are used as boundary conditions for other hyperparameters not given in the dictionary.
        Parameters:
            bounds_dict : dict
                A dictionary with boundary conditions as numpy (H,2) arrays with two columns for each type of hyperparameter.
            scale : float
                Scale the boundary conditions.
            log : bool
                Whether to use hyperparameters in log-scale or not.
            max_length : bool
                Whether to use the maximum scaling for the length-scale or use a more reasonable scaling.
            use_derivatives : bool
                Whether the derivatives of the target are used in the model. 
                The boundary conditions of the length-scale hyperparameter(s) will change with the use_derivatives. 
                The use_derivatives will be updated when update_bounds is called.
            use_prior_mean : bool
                Whether to use the prior mean to calculate the boundary of the prefactor hyperparameter.
                If use_prior_mean=False the minimum and maximum target differences are used as the boundary conditions. 
        """
        self.update_arguments(bounds_dict=bounds_dict,
                              scale=scale,
                              log=log,
                              max_length=max_length,
                              use_derivatives=use_derivatives,
                              use_prior_mean=use_prior_mean,
                              **kwargs)

    def update_arguments(self,bounds_dict=None,scale=None,log=None,max_length=None,use_derivatives=None,use_prior_mean=None,**kwargs):
        """
        Update the class with its arguments. The existing arguments are used if they are not given.
        Parameters:
            bounds_dict : dict
                A dictionary with boundary conditions as numpy (H,2) arrays with two columns for each type of hyperparameter.
            scale : float
                Scale the boundary conditions.
            log : bool
                Whether to use hyperparameters in log-scale or not.
            max_length : bool
                Whether to use the maximum scaling for the length-scale or use a more reasonable scaling.
            use_derivatives : bool
                Whether the derivatives of the target are used in the model. 
                The boundary conditions of the length-scale hyperparameter(s) will change with the use_derivatives. 
                The use_derivatives will be updated when update_bounds is called.
            use_prior_mean : bool
                Whether to use the prior mean to calculate the boundary of the prefactor hyperparameter.
                If use_prior_mean=False the minimum and maximum target differences are used as the boundary conditions. 
        Returns:
            self: The updated object itself.
        """
        if bounds_dict is not None:
            self.initiate_bounds_dict(bounds_dict)
        if scale is not None:
            self.scale=scale
        if log is not None:
            self.log=log
        if max_length is not None:
            self.max_length=max_length
        if use_derivatives is not None:
            self.use_derivatives=use_derivatives
        if use_prior_mean is not None:
            self.use_prior_mean=use_prior_mean
        return self
    
    def make_bounds(self,model,X,Y,parameters,parameters_set,**kwargs):
        " Make the boundary conditions with educated guesses of the length-scale hyperparamer(s). "
        eps_lower,eps_upper=self.get_boundary_limits()
        self.get_use_derivatives(model)
        bounds={}
        for para in parameters_set:
            if para=='length':
                bounds[para]=self.length_bound(X,parameters.count(para))
            elif para=='noise':
                if 'noise_deriv' in parameters_set:
                    bounds[para]=self.noise_bound(Y[:,0:1],eps_lower=eps_lower)
                else:
                    bounds[para]=self.noise_bound(Y,eps_lower=eps_lower)
            elif para=='noise_deriv':
                bounds[para]=self.noise_bound(Y[:,1:],eps_lower=eps_lower)
            elif para=='prefactor':
                bounds[para]=self.prefactor_bound(X,Y,model)
            elif para in self.bounds_dict:
                bounds[para]=self.bounds_dict[para].copy()
            else:
                bounds[para]=np.full((parameters.count(para),2),[eps_lower,eps_upper])
        return bounds
    
    def prefactor_bound(self,X,Y,model,**kwargs):
        " Get the minimum and maximum ranges of the prefactor in the educated guess regime within a scale. "
        if self.use_prior_mean:
            # Get the prior mean value for the target only (without derivatives)
            Y_mean=self.get_prior_mean(X,Y,model)
            Y_std=Y[:,0:1]-Y_mean
            # Calculate the variance relative to the prior mean of the targets
            a_mean=np.sqrt(np.mean(Y_std**2))
            # Check that all the targets are not the same
            if a_mean==0.0:
                a_mean=1.00
            scaling=10.0*self.scale
            a_max=a_mean*scaling
            a_min=a_mean/scaling
        else:
            # Calculate the differences in the target values
            dif=pdist(Y[:,0:1])
            # Remove zero differences
            dif=dif[dif!=0.0]
            # Check that all the targets are not the same
            if len(dif)==0:
                dif=[1.0]
            a_max=np.max(dif)*self.scale
            a_min=np.min(dif)/self.scale
        if self.log:
            return np.array([[np.log(a_min),np.log(a_max)]])
        return np.array([[a_min,a_max]])
    
    def get_prior_mean(self,X,Y,model,**kwargs):
        " Get the prior mean value for the target only (without derivatives). "
        # Update the prior mean used in ML model
        model.prior.update(X,Y)
        return model.prior.get(X,Y[:,0:1],get_derivatives=False)
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(bounds_dict=self.bounds_dict,
                        scale=self.scale,
                        log=self.log,
                        max_length=self.max_length,
                        use_derivatives=self.use_derivatives,
                        use_prior_mean=self.use_prior_mean)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs
    


import numpy as np
from .pdistributions import Prior_distribution

class Normal_prior(Prior_distribution):
    def __init__(self,mu=0.0,std=10.0,**kwargs):
        """ 
        Independent Normal prior distribution used for each type of hyperparameters in log-space. 
        If the type of the hyperparameter is multi dimensional (H) it is given in the axis=-1. 
        If multiple values (M) of the hyperparameter(/s) are calculated simultaneously it has to be in a (M,H) array. 
        Parameters:
            mu: float or (H) array
                The mean of the normal distribution. 
            std: float or (H) array
                The standard deviation of the normal distribution.
        """
        self.update_arguments(mu=mu,
                              std=std,
                              **kwargs)
    
    def ln_pdf(self,x):
        if self.nosum:
            return -np.log(self.std)-0.5*np.log(2.0*np.pi)-0.5*((x-self.mu)/self.std)**2
        return np.sum(-np.log(self.std)-0.5*np.log(2.0*np.pi)-0.5*((x-self.mu)/self.std)**2,axis=-1)
    
    def ln_deriv(self,x):
        return -(x-self.mu)/self.std**2
    
    def update_arguments(self,mu=None,std=None,**kwargs):
        """
        Update the object with its arguments. The existing arguments are used if they are not given.
        Parameters:
            mu: float or (H) array
                The mean of the normal distribution. 
            std: float or (H) array
                The standard deviation of the normal distribution.
        Returns:
            self: The updated object itself.
        """
        if mu is not None:
            if isinstance(mu,(float,int)):
                self.mu=mu
            else:
                self.mu=np.array(mu).reshape(-1)
        if std is not None:
            if isinstance(std,(float,int)):
                self.std=std
            else:
                self.std=np.array(std).reshape(-1)
        if isinstance(self.mu,(float,int)) and isinstance(self.std,(float,int)):
            self.nosum=True
        else:
            self.nosum=False
        return self
            
    def mean_var(self,mean,var):
        return self.update_arguments(mu=mean,std=np.sqrt(var))
    
    def min_max(self,min_v,max_v):
        mu=0.5*(min_v+max_v)
        return self.update_arguments(mu=mu,std=np.sqrt(2.0)*(max_v-mu))
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(mu=self.mu,std=self.std)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs



import numpy as np

class Prior_distribution:
    def __init__(self,**kwargs):
        """ 
        Prior probability distribution used for each type of hyperparameters in log-space. 
        If the type of the hyperparameter is multi dimensional (H) it is given in the axis=-1. 
        If multiple values (M) of the hyperparameter(/s) are calculated simultaneously it has to be in a (M,H) array. 
        """
        self.update_arguments(**kwargs)
        
    def pdf(self,x):
        """ 
        Probability density function.
        Parameter:
            x: float or (M,H) array
                x is the hyperparameter value used by the prior distribution.
                x can be a float if the prior distribution only consider 
                a single hyperparameter of that type.
                x can be a (1,H) array if the prior distribution consider 
                H hyperparameter of that type.  
                x can be a (M,H) array if the prior distribution consider 
                H hyperparameter of that type with M different values.  
        Returns: 
            float: Value of the probability density function.
            or
            (M) array: M values of the probability density function if M different values is given.
        """
        return np.exp(self.ln_pdf(x))
    
    def deriv(self,x):
        " The derivative of the probability density function as respect to x. "
        return self.pdf(x)*self.ln_deriv(x)
    
    def ln_pdf(self,x):
        """ 
        Log of probability density function.
        Parameter:
            x: float or (M,H) array
                x is the hyperparameter value used by the prior distribution.
                x can be a float if the prior distribution only consider 
                a single hyperparameter of that type.
                x can be a (1,H) array if the prior distribution consider 
                H hyperparameter of that type.  
                x can be a (M,H) array if the prior distribution consider 
                H hyperparameter of that type with M different values.  
        Returns: 
            float: Value of the log of probability density function.
            or
            (M) array: M values of the log of probability density function if M different values is given.
        """
        raise NotImplementedError()
    
    def ln_deriv(self,x):
        " The derivative of the log of the probability density function as respect to x. "
        raise NotImplementedError()
    
    def update_arguments(self,**kwargs):
        """
        Update the object with its arguments. The existing arguments are used if they are not given.
        Returns:
            self: The updated object itself.
        """
        return self
            
    def mean_var(self,mean,var):
        " Obtain the parameters of the distribution function by the mean and variance values. "
        raise NotImplementedError()
    
    def min_max(self,min_v,max_v):
        " Obtain the parameters of the distribution function by the minimum and maximum values. "
        raise NotImplementedError()
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict()
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs

    def copy(self):
        " Copy the object. "
        # Get all arguments
        arg_kwargs,constant_kwargs,object_kwargs=self.get_arguments()
        # Make a clone
        clone=self.__class__(**arg_kwargs)
        # Check if constants have to be saved
        if len(constant_kwargs.keys()):
            for key,value in constant_kwargs.items():
                clone.__dict__[key]=value
        # Check if objects have to be saved
        if len(object_kwargs.keys()):
            for key,value in object_kwargs.items():
                clone.__dict__[key]=value.copy()
        return clone
    
    def __repr__(self):
        arg_kwargs=self.get_arguments()[0]
        str_kwargs=",".join([f"{key}={value}" for key,value in arg_kwargs.items()])
        return "{}({})".format(self.__class__.__name__,str_kwargs)



import numpy as np

class HyperparameterFitter:
    def __init__(self,func,optimizer=None,bounds=None,use_update_pdis=False,get_prior_mean=False,use_stored_sols=False,**kwargs):
        """ 
        Hyperparameter fitter object with an optimizer for optimizing the hyperparameters on different given objective functions. 
        Parameters:
            func : ObjectiveFunction class
                A class with the objective function used to optimize the hyperparameters.
            optimizer : Optimizer class
                A class with the used optimization method.
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
                Most of the global optimizers are using boundary conditions. 
                The bounds in this class will be used for the optimizer and func.
            use_update_pdis : bool
                Whether to update the prior distributions of the hyperparameters with the given boundary conditions.
            get_prior_mean : bool
                Whether to get the parameters of the prior mean in the solution.
            use_stored_sols : bool
                Whether to store the solutions.
        """
        # Set the default optimizer
        if optimizer is None:
            from .optimizers.optimizer import FunctionEvaluation
            optimizer=FunctionEvaluation(jac=False)
        # Set the default boundary conditions
        if bounds is None:
            from .hpboundary.hptrans import VariableTransformation
            self.bounds=VariableTransformation(bounds=None)
        # Set all the arguments
        self.update_arguments(func=func,
                              optimizer=optimizer,
                              bounds=bounds,
                              use_update_pdis=use_update_pdis,
                              get_prior_mean=get_prior_mean,
                              use_stored_sols=use_stored_sols,
                              **kwargs)
        
    def fit(self,X,Y,model,hp=None,pdis=None,**kwargs):
        """ 
        Optimize the hyperparameters 
        Parameters:
            X : (N,D) array
                Training features with N data points and D dimensions.
            Y : (N,1) array or (N,D+1) array
                Training targets with or without derivatives with N data points.
            model : Model
                The Machine Learning Model with kernel and prior that are optimized.
            hp : dict
                Use a set of hyperparameters to optimize from else the current set is used.
            pdis : dict
                A dict of prior distributions for each hyperparameter type.
        Returns: 
            dict : A solution dictionary with objective function value, optimized hyperparameters,
                success statement, and number of used evaluations.
        """
        # Copy the model so it is not changed outside of the optimization
        model=self.copy_model(model)
        # Always reset the solution in the objective function
        self.reset_func()
        # Get hyperparameters
        hp,theta,parameters=self.get_hyperparams(hp,model)
        # Update bounds
        self.update_bounds(model,X,Y,parameters)
        # Update prior distributions of hyperparameters
        pdis=self.update_pdis(pdis,model,X,Y,parameters)
        # Modify the hyperparameters
        theta,parameters=self.modify_hyperparams(hp)
        # Optimize the hyperparameters
        sol=self.optimizer.run(self.func,theta,parameters,model,X,Y,pdis=pdis)
        # Get the full set of hyperparameters in the model
        sol=self.get_full_hp(sol,model)
        # Store the solution 
        self.store_sol(sol)
        return sol
    
    def update_arguments(self,func=None,optimizer=None,bounds=None,use_update_pdis=None,get_prior_mean=None,use_stored_sols=None,**kwargs):
        """
        Update the class with its arguments. The existing arguments are used if they are not given.
        Parameters:
            func : ObjectiveFunction class
                A class with the objective function used to optimize the hyperparameters.
            optimizer : Optimizer class
                A class with the used optimization method.
            bounds : HPBoundaries class
                A class of the boundary conditions of the hyperparameters.
                Most of the global optimizers are using boundary conditions. 
                The bounds in this class will be used for the optimizer and func.
            use_update_pdis : bool
                Whether to update the prior distributions of the hyperparameters with the given boundary conditions.
            get_prior_mean : bool
                Whether to get the parameters of the prior mean in the solution.
            use_stored_sols : bool
                Whether to store the solutions.
        Returns:
            self: The updated object itself.
        """
        if func is not None:
            self.func=func.copy()
        if optimizer is not None:
            self.optimizer=optimizer.copy()
        if bounds is not None:
            self.bounds=bounds.copy()
        if use_update_pdis is not None:
            self.use_update_pdis=use_update_pdis
        if get_prior_mean is not None:
            self.get_prior_mean=get_prior_mean
        if use_stored_sols is not None:
            self.use_stored_sols=use_stored_sols
        # Empty the stored solutions
        self.sols=[]
        # Make sure that the objective function gets the prior mean parameters or not
        self.func.update_arguments(get_prior_mean=self.get_prior_mean)
        return self

    def copy_model(self,model,**kwargs):
        " Make a copy of the model, so it is not overwritten. "
        return model.copy()
    
    def reset_func(self,**kwargs):
        " Reset the solution in objective function object. "
        return self.func.reset_solution()
    
    def get_hyperparams(self,hp,model,**kwargs):
        " Get default hyperparameters if they are not given. "
        if hp is None:
            # Get the hyperparameters from the model
            hp=model.get_hyperparams()
        # Get the values and hyperparameter names
        theta,parameters=self.hp_to_theta(hp)
        return hp,theta,parameters
    
    def modify_hyperparams(self,hp,**kwargs):
        " Modify the hyperparameters by checking what parameters is used by the objective function. "
        if self.func.use_analytic_prefactor:
            # Remove the prefactor hyperparameter if it is solved internally 
            hp.pop('prefactor',None)
        if self.func.use_optimized_noise:
            # Remove the relative-noise hyperparameter if it is solved internally 
            hp.pop('noise',None)
        theta,parameters=self.hp_to_theta(hp)
        return theta,parameters
    
    def hp_to_theta(self,hp):
        " Transform a dictionary of hyperparameters to a list of values and a list of hyperparameter names. " 
        parameters_set=sorted(hp.keys())
        theta=sum([list(hp[para]) for para in parameters_set],[])
        parameters=sum([[para]*len(hp[para]) for para in parameters_set],[])
        return np.array(theta),parameters
    
    def update_bounds(self,model,X,Y,parameters,**kwargs):
        " Update the boundary condition class with the data. "
        # Give the stored solutions to the bounds if used
        if self.use_stored_sols:
            self.bounds.update_arguments(sols=self.get_sols())
        # Update the boundary conditions
        self.bounds.update_bounds(model,X,Y,parameters)
        # Update the bounds in the objective function
        self.func.update_arguments(bounds=self.bounds)
        # Update the bounds in the optimizer object
        self.optimizer.update_arguments(bounds=self.bounds)
        return self.bounds
    
    def update_pdis(self,pdis,model,X,Y,parameters,**kwargs):
        " Update the prior distributions of the hyperparameters with the boundary conditions. "
        if self.use_update_pdis and pdis is not None:
            from .pdistributions.update_pdis import update_pdis
            pdis=update_pdis(model,parameters,X,Y,bounds=self.bounds,pdis=pdis)
        return pdis
    
    def get_full_hp(self,sol,model,**kwargs):
        " Get the full hyperparameter dictionary with hyperparameters that are optimized and within the model. "
        sol['full hp']=model.get_hyperparams()
        sol['full hp'].update(sol['hp'])
        return sol

    def store_sol(self,sol,**kwargs):
        " Store the solutions. "
        if self.use_stored_sols:
            self.sols.append(sol)
        return self.sols

    def get_sols(self,**kwargs):
        " Get the stored solutions. "
        return self.sols.copy()
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(func=self.func,
                        optimizer=self.optimizer,
                        bounds=self.bounds,
                        use_update_pdis=self.use_update_pdis,
                        get_prior_mean=self.get_prior_mean,
                        use_stored_sols=self.use_stored_sols)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict(sols=self.get_sols())
        return arg_kwargs,constant_kwargs,object_kwargs
    
    def copy(self):
        " Copy the object. "
        # Get all arguments
        arg_kwargs,constant_kwargs,object_kwargs=self.get_arguments()
        # Make a clone
        clone=self.__class__(**arg_kwargs)
        # Check if constants have to be saved
        if len(constant_kwargs.keys()):
            for key,value in constant_kwargs.items():
                clone.__dict__[key]=value
        # Check if objects have to be saved
        if len(object_kwargs.keys()):
            for key,value in object_kwargs.items():
                clone.__dict__[key]=value.copy()
        return clone
    
    def __repr__(self):
        arg_kwargs=self.get_arguments()[0]
        str_kwargs=",".join([f"{key}={value}" for key,value in arg_kwargs.items()])
        return "{}({})".format(self.__class__.__name__,str_kwargs)
    


import numpy as np
from scipy.linalg import cho_solve
from .objectivefunction_gpatom import ObjectiveFuctionGPAtom

class MaximumLogLikelihood(ObjectiveFuctionGPAtom):
    def __init__(self,get_prior_mean=False,modification=False,**kwargs):
        """ 
        The Maximum log-likelihood objective function as a function of the hyperparameters. 
        The prefactor hyperparameter is calculated from an analytical expression. 
        Parameters:
            get_prior_mean: bool
                Whether to save the parameters of the prior mean in the solution.
            modification: bool
                Whether to modify the analytical prefactor value in the end.
                The prefactor hyperparameter becomes larger if modification=True.
        """
        # Set descriptor of the objective function
        self.use_analytic_prefactor=True
        self.use_optimized_noise=False
        # Set the arguments
        self.update_arguments(get_prior_mean=get_prior_mean,
                              modification=modification,
                              **kwargs)
    
    def function(self,theta,parameters,model,X,Y,pdis=None,jac=False,**kwargs):
        hp,parameters_set=self.make_hp(theta,parameters)
        model=self.update_model(model,hp)
        coef,L,low,Y_p,KXX,n_data=self.coef_cholesky(model,X,Y)
        prefactor2=np.matmul(Y_p.T,coef).item(0)/n_data
        prefactor=0.5*np.log(prefactor2)
        hp['prefactor']=np.array([prefactor])
        nlp=0.5*n_data*(1+np.log(2.0*np.pi))+n_data*prefactor+np.sum(np.log(np.diagonal(L)))
        nlp=nlp-self.logpriors(hp,pdis,jac=False)
        if jac:
            deriv=self.derivative(hp,parameters_set,model,X,KXX,L,low,coef,prefactor2,n_data,pdis,**kwargs) 
            self.update_solution(nlp,theta,hp,model,jac=jac,deriv=deriv,prefactor2=prefactor2,n_data=n_data)
            return nlp,deriv
        self.update_solution(nlp,theta,hp,model,jac=jac,prefactor2=prefactor2,n_data=n_data)
        return nlp
    
    def derivative(self,hp,parameters_set,model,X,KXX,L,low,coef,prefactor2,n_data,pdis,**kwargs):
        nlp_deriv=np.array([])
        KXX_inv=cho_solve((L,low),np.identity(n_data),check_finite=False)
        for para in parameters_set:
            if para=='prefactor':
                nlp_deriv=np.append(nlp_deriv,np.zeros((len(hp[para]))))
                continue
            K_deriv=self.get_K_deriv(model,para,X=X,KXX=KXX)
            K_deriv_cho=self.get_K_inv_deriv(K_deriv,KXX_inv)
            nlp_deriv=np.append(nlp_deriv,(-(0.5*np.matmul(coef.T,np.matmul(K_deriv,coef)).reshape(-1))/prefactor2)+0.5*K_deriv_cho)
        nlp_deriv=nlp_deriv-self.logpriors(hp,pdis,jac=True)
        return nlp_deriv
    
    def update_arguments(self,get_prior_mean=None,modification=None,**kwargs):
        """
        Update the objective function with its arguments. The existing arguments are used if they are not given.
        Parameters:
            get_prior_mean : bool
                Whether to get the parameters of the prior mean in the solution.
            modification: bool
                Whether to modify the analytical prefactor value in the end.
                The prefactor hyperparameter becomes larger if modification=True.
        Returns:
            self: The updated object itself.
        """
        if get_prior_mean is not None:
            self.get_prior_mean=get_prior_mean
        if modification is not None:
            self.modification=modification
        # Always reset the solution when the objective function is changed 
        self.reset_solution()
        return self
    
    def update_solution(self,fun,theta,hp,model,jac=False,deriv=None,prefactor2=None,n_data=None,**kwargs):
        """
        Update the solution of the optimization in terms of hyperparameters and model.
        The lowest objective function value is stored togeher with its hyperparameters.
        The prior mean can also be saved if get_prior_mean=True.
        The prefactor hyperparameter are stored as a different value
        than the input since it is optimized analytically.
        """
        if fun<self.sol['fun']:
            if self.modification:
                prefactor2=(n_data/(n_data-len(theta)))*prefactor2 if n_data-len(theta)>0 else prefactor2
                hp['prefactor']=np.array([0.5*np.log(prefactor2)])
            self.sol['x']=np.concatenate([hp[para] for para in sorted(hp.keys())])
            self.sol['hp']=hp.copy()
            self.sol['fun']=fun
            if jac:
                self.sol['jac']=deriv.copy()
            if self.get_prior_mean:
                self.sol['prior']=self.get_prior_parameters(model)
        return self.sol
    
    def get_arguments(self):
        " Get the arguments of the class itself. "
        # Get the arguments given to the class in the initialization
        arg_kwargs=dict(get_prior_mean=self.get_prior_mean,
                        modification=self.modification)
        # Get the constants made within the class
        constant_kwargs=dict()
        # Get the objects made within the class
        object_kwargs=dict()
        return arg_kwargs,constant_kwargs,object_kwargs
